"""Tests for linear constraints and objectives."""
import numpy as np
import pytest
import scipy.linalg
from qsc import Qsc

import desc.examples
from desc.compute import arg_order
from desc.equilibrium import Equilibrium
from desc.geometry import FourierRZToroidalSurface
from desc.grid import LinearGrid
from desc.io import load
from desc.objectives import (
    AspectRatio,
    AxisRSelfConsistency,
    AxisZSelfConsistency,
    BoundaryRSelfConsistency,
    BoundaryZSelfConsistency,
    FixAtomicNumber,
    FixAxisR,
    FixAxisZ,
    FixBoundaryR,
    FixBoundaryZ,
    FixCurrent,
    FixElectronDensity,
    FixElectronTemperature,
    FixIonTemperature,
    FixIota,
    FixLambdaGauge,
    FixModeLambda,
    FixModeR,
    FixModeZ,
    FixPressure,
    FixPsi,
    FixSumModesLambda,
    FixSumModesR,
    FixSumModesZ,
    FixThetaSFL,
    ObjectiveFunction,
    QuasisymmetryTwoTerm,
    get_fixed_axis_constraints,
    get_fixed_boundary_constraints,
    get_NAE_constraints,
)
from desc.profiles import PowerSeriesProfile

# TODO: check for all bdryR things if work when False is passed in
# bc empty array indexed will lead to an error


@pytest.mark.unit
def test_LambdaGauge_sym(DummyStellarator):
    """Test that lambda is fixed correctly for symmetric equilibrium."""
    # symmetric cases automatically satisfy gauge freedom, no constraint needed.
    eq = load(load_from=str(DummyStellarator["output_path"]), file_format="hdf5")
    eq.change_resolution(L=2, M=1, N=1)
    correct_constraint_matrix = np.zeros((0, 5))
    lam_con = FixLambdaGauge(eq)
    lam_con.build()
    np.testing.assert_array_equal(lam_con._A, correct_constraint_matrix)


@pytest.mark.unit
def test_LambdaGauge_asym():
    """Test that lambda gauge is fixed correctly for asymmetric equilibrium."""
    # just testing the gauge condition
    inputs = {
        "sym": False,
        "NFP": 3,
        "Psi": 1.0,
        "L": 2,
        "M": 2,
        "N": 1,
        "pressure": np.array([[0, 1e4], [2, -2e4], [4, 1e4]]),
        "iota": np.array([[0, 0.5], [2, 0.5]]),
        "surface": np.array(
            [
                [0, 0, 0, 3, 0],
                [0, 1, 0, 1, 0],
                [0, -1, 0, 0, -1],
                [0, 1, 1, 0.3, 0],
                [0, -1, -1, 0.3, 0],
                [0, 1, -1, 0, -0.3],
                [0, -1, 1, 0, 0.3],
            ],
        ),
        "axis": np.array([[-1, 0, -0.2], [0, 3.4, 0], [1, 0.2, 0]]),
        "objective": "force",
        "optimizer": "lsq-exact",
    }
    eq = Equilibrium(**inputs)
    lam_con = FixLambdaGauge(eq)
    lam_con.build()

    # make sure that any lambda in the null space gives lambda==0 at theta=zeta=0
    Z = scipy.linalg.null_space(lam_con._A)
    grid = LinearGrid(L=10, theta=[0], zeta=[0])
    for z in Z.T:
        eq.L_lmn = z
        lam = eq.compute("lambda", grid=grid)["lambda"]
        np.testing.assert_allclose(lam, 0, atol=1e-15)


@pytest.mark.unit
def test_bc_on_interior_surfaces():
    """Test applying boundary conditions on internal surface."""
    surf = FourierRZToroidalSurface(rho=0.5)
    iota = PowerSeriesProfile([1, 0, 0.5])
    eq = Equilibrium(L=4, M=4, N=0, surface=surf, iota=iota)
    eq.solve(verbose=0)
    surf5 = eq.get_surface_at(0.5)

    np.testing.assert_allclose(surf.R_lmn, surf5.R_lmn, atol=1e-12)
    np.testing.assert_allclose(surf.Z_lmn, surf5.Z_lmn, atol=1e-12)


@pytest.mark.unit
def test_constrain_bdry_with_only_one_mode():
    """Test Fixing boundary with a surface with only one mode in its basis."""
    eq = Equilibrium()
    FixZ = BoundaryZSelfConsistency(eq=eq)
    try:
        FixZ.build()
    except Exception:
        pytest.fail(
            "Error encountered when attempting to constrain surface with"
            + " only one mode in its basis"
        )


@pytest.mark.unit
def test_constrain_asserts():
    """Test error checking for incompatible constraints."""
    eqi = Equilibrium(iota=PowerSeriesProfile(0, 0), pressure=PowerSeriesProfile(0, 0))
    eqc = Equilibrium(current=PowerSeriesProfile(0))
    # nonexistent toroidal current can't be constrained
    with pytest.raises(RuntimeError):
        eqi.solve(constraints=FixCurrent(eq=eqi))
    # nonexistent rotational transform can't be constrained
    with pytest.raises(RuntimeError):
        eqc.solve(constraints=FixIota(eq=eqc))
    # toroidal current and rotational transform can't be constrained simultaneously
    with pytest.raises(ValueError):
        eqi.solve(constraints=(FixCurrent(eq=eqi), FixIota(eq=eqi)))
    with pytest.raises(AssertionError):
        eqi.solve(
            constraints=(FixPressure(eq=eqi, target=2), FixPressure(eq=eqi, target=1))
        )
    # cannot use two incompatible constraints
    with pytest.raises(AssertionError):
        con1 = FixCurrent(target=eqc.c_l, eq=eqc)
        con2 = FixCurrent(target=eqc.c_l + 1, eq=eqc)
        eqc.solve(constraints=(con1, con2))


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_fixed_mode_solve():
    """Test solving an equilibrium with a fixed mode constraint."""
    # Reset DSHAPE to initial guess, fix a mode, and then resolve
    # and check that the mode stayed fix
    L = 1
    M = 1
    eq = desc.examples.get("DSHAPE")
    eq.set_initial_guess()
    fixR = FixModeR(
        eq=eq, modes=np.array([L, M, 0])
    )  # no target supplied, so defaults to the eq's current R_LMN coeff
    fixZ = FixModeZ(
        eq=eq, modes=np.array([L, -M, 0])
    )  # no target supplied, so defaults to the eq's current Z_LMN coeff
    orig_R_val = eq.R_lmn[eq.R_basis.get_idx(L=L, M=M, N=0)]
    orig_Z_val = eq.Z_lmn[eq.Z_basis.get_idx(L=L, M=-M, N=0)]

    constraints = (
        FixLambdaGauge(eq=eq),
        FixPressure(eq=eq),
        FixIota(eq=eq),
        FixPsi(eq=eq),
        FixBoundaryR(eq=eq),
        FixBoundaryZ(eq=eq),
        fixR,
        fixZ,
    )

    eq.solve(
        verbose=3,
        ftol=1e-2,
        objective="force",
        maxiter=10,
        xtol=1e-6,
        constraints=constraints,
    )
    np.testing.assert_almost_equal(
        orig_R_val, eq.R_lmn[eq.R_basis.get_idx(L=L, M=M, N=0)]
    )
    np.testing.assert_almost_equal(
        orig_Z_val, eq.Z_lmn[eq.Z_basis.get_idx(L=L, M=-M, N=0)]
    )


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_fixed_modes_solve():
    """Test solving an equilibrium with fixed sum modes constraint."""
    # Reset DSHAPE to initial guess, fix sum modes, and then resolve
    # and check that the mode sum stayed fix
    modes_R = np.array([[1, 1, 0], [2, 2, 0]])
    modes_Z = np.array([[1, -1, 0], [2, -2, 0]])

    eq = desc.examples.get("DSHAPE")
    eq.set_initial_guess()
    fixR = FixSumModesR(
        eq=eq, modes=modes_R, sum_weights=np.array([1, 2])
    )  # no target supplied, so defaults to the eq's current sum
    fixZ = FixSumModesZ(
        eq=eq, modes=modes_Z, sum_weights=np.array([1, 2])
    )  # no target supplied, so defaults to the eq's current sum
    fixLambda = FixSumModesLambda(
        eq=eq, modes=modes_Z, sum_weights=np.array([1, 2])
    )  # no target supplied, so defaults to the eq's current sum

    orig_R_val = (
        eq.R_lmn[eq.R_basis.get_idx(L=modes_R[0, 0], M=modes_R[0, 1], N=0)]
        + 2 * eq.R_lmn[eq.R_basis.get_idx(L=modes_R[1, 0], M=modes_R[1, 1], N=0)]
    )
    orig_Z_val = (
        eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )
    orig_Lambda_val = (
        eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )
    constraints = (
        FixLambdaGauge(eq=eq),
        FixPressure(eq=eq),
        FixIota(eq=eq),
        FixPsi(eq=eq),
        FixBoundaryR(eq=eq),
        FixBoundaryZ(eq=eq),
        fixR,
        fixZ,
        fixLambda,
    )

    eq.solve(
        verbose=3,
        ftol=1e-2,
        objective="force",
        maxiter=10,
        xtol=1e-6,
        constraints=constraints,
    )

    new_R_val = (
        eq.R_lmn[eq.R_basis.get_idx(L=modes_R[0, 0], M=modes_R[0, 1], N=0)]
        + 2 * eq.R_lmn[eq.R_basis.get_idx(L=modes_R[1, 0], M=modes_R[1, 1], N=0)]
    )
    new_Z_val = (
        eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.Z_lmn[eq.Z_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )
    new_Lambda_val = (
        eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[0, 0], M=modes_Z[0, 1], N=0)]
        + 2 * eq.L_lmn[eq.L_basis.get_idx(L=modes_Z[1, 0], M=modes_Z[1, 1], N=0)]
    )

    np.testing.assert_almost_equal(orig_R_val, new_R_val)
    np.testing.assert_almost_equal(orig_Z_val, new_Z_val)
    np.testing.assert_almost_equal(orig_Lambda_val, new_Lambda_val)


@pytest.mark.regression
@pytest.mark.solve
@pytest.mark.slow
def test_fixed_axis_and_theta_SFL_solve():
    """Test solving an equilibrium with a fixed axis and theta SFL constraint."""
    # also tests zero lambda solve
    eq = Equilibrium()

    orig_R_val = eq.axis.R_n
    orig_Z_val = eq.axis.Z_n

    constraints = (
        FixThetaSFL(eq=eq),
        FixPressure(eq=eq),
        FixCurrent(eq=eq),
        FixPsi(eq=eq),
        FixAxisR(eq=eq),
        FixAxisZ(eq=eq),
    )

    eq.solve(
        verbose=3,
        ftol=1e-2,
        objective="force",
        maxiter=10,
        xtol=1e-6,
        constraints=constraints,
    )

    np.testing.assert_allclose(orig_R_val, eq.axis.R_n, atol=1e-14)
    np.testing.assert_allclose(orig_Z_val, eq.axis.Z_n, atol=1e-14)
    np.testing.assert_allclose(np.zeros_like(eq.L_lmn), eq.L_lmn, atol=1e-14)


@pytest.mark.unit
def test_factorize_linear_constraints_asserts():
    """Test error checking for factorize_linear_constraints."""
    eq = Equilibrium()
    constraints = get_fixed_boundary_constraints(eq=eq, iota=False)
    for con in constraints:
        con.build(verbose=0)
    constraints[3].bounds = (0, 1)  # bounds on FixPsi

    from desc.objectives.utils import factorize_linear_constraints

    with pytest.raises(ValueError):
        xp, A, b, Z, unfixed_idx, project, recover = factorize_linear_constraints(
            constraints, arg_order
        )


@pytest.mark.unit
def test_build_init():
    """Ensure that passing an equilibrium to init builds the objective correctly.

    Related to gh issue #378
    """
    eq = Equilibrium(M=3, N=1)

    # initialize the constraints without building
    fbR1 = FixBoundaryR(eq=eq)
    fbZ1 = FixBoundaryZ(eq=eq)
    for obj in (fbR1, fbZ1):
        obj.build()

    arg = fbR1.args[0]
    A = fbR1.derivatives["jac_scaled"][arg](np.zeros(fbR1.dimensions[arg]))
    assert np.max(np.abs(A)) == 1
    assert A.shape == (eq.surface.R_basis.num_modes, eq.surface.R_basis.num_modes)

    arg = fbZ1.args[0]
    A = fbZ1.derivatives["jac_scaled"][arg](np.zeros(fbZ1.dimensions[arg]))
    assert np.max(np.abs(A)) == 1
    assert A.shape == (eq.surface.Z_basis.num_modes, eq.surface.Z_basis.num_modes)

    fbR1 = FixBoundaryR(eq=eq)
    fbZ1 = FixBoundaryZ(eq=eq)
    for obj in (fbR1, fbZ1):
        obj.build()

    arg = fbR1.args[0]
    A = fbR1.derivatives["jac_scaled"][arg](np.zeros(fbR1.dimensions[arg]))
    assert np.max(np.abs(A)) == 1
    assert A.shape == (eq.surface.R_basis.num_modes, eq.surface.R_basis.num_modes)

    arg = fbZ1.args[0]
    A = fbZ1.derivatives["jac_scaled"][arg](np.zeros(fbZ1.dimensions[arg]))
    assert np.max(np.abs(A)) == 1
    assert A.shape == (eq.surface.Z_basis.num_modes, eq.surface.Z_basis.num_modes)


@pytest.mark.unit
def test_kinetic_constraints():
    """Make sure errors are raised when trying to constrain nonexistent profiles."""
    eqp = Equilibrium(L=3, M=3, N=3, pressure=np.array([1, 0, -1]))
    eqk = Equilibrium(
        L=3,
        M=3,
        N=3,
        electron_temperature=np.array([1, 0, -1]),
        electron_density=np.array([2, 0, -2]),
    )
    pcon = (FixPressure(eq=eqk),)
    kcon = (
        FixAtomicNumber(eq=eqp),
        FixElectronDensity(eq=eqp),
        FixElectronTemperature(eq=eqp),
        FixIonTemperature(eq=eqp),
    )
    for con in pcon:
        with pytest.raises(RuntimeError):
            con.build()
    for con in kcon:
        with pytest.raises(RuntimeError):
            con.build()


@pytest.mark.unit
def test_correct_indexing_passed_modes():
    """Test Indexing when passing in specified modes, related to gh issue #380."""
    n = 1

    eq = desc.examples.get("W7-X")

    grid = LinearGrid(
        M=eq.M, N=eq.N, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True
    )

    objective = ObjectiveFunction(
        (
            QuasisymmetryTwoTerm(eq=eq, weight=1e-2, helicity=(1, -eq.NFP), grid=grid),
            AspectRatio(eq=eq, target=8, weight=1e2),
        ),
        verbose=0,
    )
    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > n + 1, :
            ],
        )
    )
    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > n + 1, :
    ]
    constraints = (
        FixBoundaryR(eq=eq, modes=R_modes, normalize=False),
        FixBoundaryZ(eq=eq, modes=Z_modes, normalize=False),
        BoundaryRSelfConsistency(eq=eq),
        BoundaryZSelfConsistency(eq=eq),
    )
    for con in constraints:
        con.build(verbose=0)
    objective.build()
    objective.set_args("Rb_lmn", "Zb_lmn")
    from desc.objectives.utils import factorize_linear_constraints

    xp, A, b, Z, unfixed_idx, project, recover = factorize_linear_constraints(
        constraints,
        objective.args,
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    assert np.isclose(np.max(np.abs(x1 - x2)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ xp[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x1[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x2[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ Z)), 0, atol=atol)


@pytest.mark.unit
def test_correct_indexing_passed_modes_and_passed_target():
    """Test Indexing when passing in specified modes, related to gh issue #380."""
    n = 1

    eq = desc.examples.get("W7-X")

    grid = LinearGrid(
        M=eq.M, N=eq.N, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True
    )

    objective = ObjectiveFunction(
        (
            QuasisymmetryTwoTerm(eq=eq, weight=1e-2, helicity=(1, -eq.NFP), grid=grid),
            AspectRatio(eq=eq, target=8, weight=1e2),
        ),
        verbose=0,
    )
    R_modes = np.vstack(
        (
            [0, 0, 0],
            eq.surface.R_basis.modes[
                np.max(np.abs(eq.surface.R_basis.modes), 1) > n + 1, :
            ],
        )
    )
    idxs = []
    for mode in R_modes:
        idxs.append(eq.surface.R_basis.get_idx(*mode))
    target_R = eq.surface.R_lmn[idxs]

    Z_modes = eq.surface.Z_basis.modes[
        np.max(np.abs(eq.surface.Z_basis.modes), 1) > n + 1, :
    ]
    idxs = []
    for mode in Z_modes:
        idxs.append(eq.surface.Z_basis.get_idx(*mode))
    target_Z = eq.surface.Z_lmn[idxs]
    constraints = (
        FixBoundaryR(eq=eq, modes=R_modes, normalize=False, target=target_R),
        FixBoundaryZ(eq=eq, modes=Z_modes, normalize=False, target=target_Z),
        BoundaryRSelfConsistency(eq=eq),
        BoundaryZSelfConsistency(eq=eq),
    )
    for con in constraints:
        con.build(eq, verbose=0)
    objective.build(eq)
    objective.set_args("Rb_lmn", "Zb_lmn")
    from desc.objectives.utils import factorize_linear_constraints

    xp, A, b, Z, unfixed_idx, project, recover = factorize_linear_constraints(
        constraints,
        objective.args,
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    assert np.isclose(np.max(np.abs(x1 - x2)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ xp[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x1[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x2[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ Z)), 0, atol=atol)


@pytest.mark.unit
def test_correct_indexing_passed_modes_axis():
    """Test Indexing when passing in specified axis modes, related to gh issue #380."""
    n = 1

    eq = desc.examples.get("W7-X")

    grid = LinearGrid(
        M=eq.M, N=eq.N, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True
    )

    objective = ObjectiveFunction(
        (
            QuasisymmetryTwoTerm(eq=eq, weight=1e-2, helicity=(1, -eq.NFP), grid=grid),
            AspectRatio(eq=eq, target=8, weight=1e2),
        ),
        verbose=0,
    )
    R_modes = np.vstack(
        (
            eq.axis.R_basis.modes[np.max(np.abs(eq.axis.R_basis.modes), 1) > n + 1, :],
            [0, 0, 0],
        )
    )
    R_modes = np.flip(R_modes, 0)

    Z_modes = eq.axis.Z_basis.modes[np.max(np.abs(eq.axis.Z_basis.modes), 1) > n + 1, :]
    Z_modes = np.flip(Z_modes, 0)

    constraints = (
        FixAxisR(eq=eq, modes=R_modes, normalize=False),
        FixAxisZ(eq=eq, modes=Z_modes, normalize=False),
        AxisRSelfConsistency(eq=eq),
        AxisZSelfConsistency(eq=eq),
        FixModeR(eq=eq, modes=np.array([[1, 1, 1], [2, 2, 2]]), normalize=False),
        FixModeZ(eq=eq, modes=np.array([[1, 1, -1], [2, 2, -2]]), normalize=False),
        FixModeLambda(eq=eq, modes=np.array([[1, 1, -1], [2, 2, -2]]), normalize=False),
        FixSumModesR(
            eq=eq,
            modes=np.array([[3, 3, 3], [4, 4, 4]]),
            normalize=False,
            sum_weights=np.ones(2),
        ),
        FixSumModesZ(eq=eq, modes=np.array([[3, 3, -3], [4, 4, -4]]), normalize=False),
    )
    for con in constraints:
        con.build(verbose=0)
    objective.build()
    objective.set_args("Ra_n", "Za_n")
    from desc.objectives.utils import factorize_linear_constraints

    xp, A, b, Z, unfixed_idx, project, recover = factorize_linear_constraints(
        constraints,
        objective.args,
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    assert np.isclose(np.max(np.abs(x1 - x2)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ xp[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x1[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x2[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ Z)), 0, atol=atol)


@pytest.mark.unit
def test_correct_indexing_passed_modes_and_passed_target_axis():
    """Test Indexing when passing in specified axis modes, related to gh issue #380."""
    n = 1

    eq = desc.examples.get("W7-X")

    grid = LinearGrid(
        M=eq.M, N=eq.N, NFP=eq.NFP, rho=np.array([0.6, 0.8, 1.0]), sym=True
    )

    objective = ObjectiveFunction(
        (
            QuasisymmetryTwoTerm(eq=eq, weight=1e-2, helicity=(1, -eq.NFP), grid=grid),
            AspectRatio(eq=eq, target=8, weight=1e2),
        ),
        verbose=0,
    )
    R_modes = np.vstack(
        (
            eq.axis.R_basis.modes[np.max(np.abs(eq.axis.R_basis.modes), 1) > n + 1, :],
            [0, 0, 0],
        )
    )
    R_modes = np.flip(R_modes, 0)
    idxs = []
    for mode in R_modes:
        idxs.append(eq.axis.R_basis.get_idx(*mode))
    target_R = eq.axis.R_n[idxs]

    Z_modes = eq.axis.Z_basis.modes[np.max(np.abs(eq.axis.Z_basis.modes), 1) > n + 1, :]
    Z_modes = np.flip(Z_modes, 0)
    idxs = []
    for mode in Z_modes:
        idxs.append(eq.axis.Z_basis.get_idx(*mode))
    target_Z = eq.axis.Z_n[idxs]

    constraints = (
        AxisRSelfConsistency(eq=eq),
        AxisZSelfConsistency(eq=eq),
        FixAxisR(eq=eq, modes=R_modes, normalize=False, target=target_R),
        FixAxisZ(eq=eq, modes=Z_modes, normalize=False, target=target_Z),
        FixModeR(
            eq=eq,
            modes=np.array([[1, 1, 1], [2, 2, 2]]),
            target=np.array(
                [
                    eq.R_lmn[eq.R_basis.get_idx(*(1, 1, 1))],
                    eq.R_lmn[eq.R_basis.get_idx(*(2, 2, 2))],
                ]
            ),
            normalize=False,
        ),
        FixModeZ(
            eq=eq,
            modes=np.array([[1, 1, -1], [2, 2, -2]]),
            target=np.array(
                [
                    eq.Z_lmn[eq.Z_basis.get_idx(*(1, 1, -1))],
                    eq.Z_lmn[eq.Z_basis.get_idx(*(2, 2, -2))],
                ]
            ),
            normalize=False,
        ),
        FixModeLambda(
            eq=eq,
            modes=np.array([[1, 1, -1], [2, 2, -2]]),
            target=np.array(
                [
                    eq.L_lmn[eq.L_basis.get_idx(*(1, 1, -1))],
                    eq.L_lmn[eq.L_basis.get_idx(*(2, 2, -2))],
                ]
            ),
            normalize=False,
        ),
        FixSumModesR(
            eq=eq,
            modes=np.array([[3, 3, 3], [4, 4, 4]]),
            target=np.array(
                [
                    eq.R_lmn[eq.R_basis.get_idx(*(3, 3, 3))]
                    + eq.R_lmn[eq.R_basis.get_idx(*(4, 4, 4))]
                ]
            ),
            normalize=False,
        ),
        FixSumModesZ(
            eq=eq,
            modes=np.array([[3, 3, -3], [4, 4, -4]]),
            target=np.array(
                [
                    eq.Z_lmn[eq.Z_basis.get_idx(*(3, 3, -3))]
                    + eq.Z_lmn[eq.Z_basis.get_idx(*(4, 4, -4))]
                ]
            ),
            normalize=False,
        ),
        FixSumModesLambda(
            eq=eq,
            modes=np.array([[3, 3, -3], [4, 4, -4]]),
            target=np.array(
                [
                    eq.L_lmn[eq.L_basis.get_idx(*(3, 3, -3))]
                    + eq.L_lmn[eq.L_basis.get_idx(*(4, 4, -4))]
                ]
            ),
            normalize=False,
        ),
    )
    for con in constraints:
        con.build(verbose=0)
    objective.build()
    objective.set_args("Ra_n", "Za_n")
    from desc.objectives.utils import factorize_linear_constraints

    xp, A, b, Z, unfixed_idx, project, recover = factorize_linear_constraints(
        constraints,
        objective.args,
    )

    x1 = objective.x(eq)
    x2 = recover(project(x1))

    atol = 2e-15
    assert np.isclose(np.max(np.abs(x1 - x2)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ xp[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x1[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ x2[unfixed_idx] - b)), 0, atol=atol)
    assert np.isclose(np.max(np.abs(A @ Z)), 0, atol=atol)


@pytest.mark.unit
def test_FixBoundary_with_single_weight():
    """Test Fixing boundary with only a single, passed weight."""
    eq = Equilibrium()
    w = 1.1
    FixZ = FixBoundaryZ(eq=eq, modes=np.array([[0, -1, 0]]), weight=w)
    FixZ.build()
    np.testing.assert_array_equal(FixZ.weight.size, 1)
    np.testing.assert_array_equal(FixZ.weight, w)
    FixR = FixBoundaryR(eq=eq, modes=np.array([[0, 1, 0]]), weight=w)
    FixR.build()
    np.testing.assert_array_equal(FixR.weight.size, 1)
    np.testing.assert_array_equal(FixR.weight, w)


@pytest.mark.unit
def test_FixBoundary_passed_target_no_passed_modes_error():
    """Test Fixing boundary with no passed-in modes."""
    eq = Equilibrium()
    FixZ = FixBoundaryZ(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixZ.build()
    FixZ = FixBoundaryZ(eq=eq, modes=False, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixZ.build()
    FixR = FixBoundaryR(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixR.build()
    FixR = FixBoundaryR(eq=eq, modes=False, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixR.build()


@pytest.mark.unit
def test_FixAxis_passed_target_no_passed_modes_error():
    """Test Fixing Axis with no passed-in modes."""
    eq = Equilibrium()
    FixZ = FixAxisZ(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixZ.build()
    FixZ = FixAxisZ(eq=eq, modes=False, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixZ.build()
    FixR = FixAxisR(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixR.build()
    FixR = FixAxisR(eq=eq, modes=False, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixR.build()


@pytest.mark.unit
def test_FixMode_passed_target_no_passed_modes_error():
    """Test Fixing Modes with no passed-in modes."""
    eq = Equilibrium()
    FixZ = FixModeZ(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixZ.build()
    FixR = FixModeR(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixR.build(eq)
    FixL = FixModeLambda(eq=eq, modes=True, target=np.array([[0]]))
    with pytest.raises(RuntimeError):
        FixL.build(eq)


@pytest.mark.unit
def test_FixSumModes_passed_target_too_long():
    """Test Fixing Modes with more than a size-1 target."""
    # TODO: remove this test if FixSumModes is generalized
    # to accept multiple targets and sets of modes at a time
    with pytest.raises(ValueError):
        FixSumModesZ(modes=np.array([[0, 0, 0], [1, 1, 1]]), target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesR(modes=np.array([[0, 0, 0], [1, 1, 1]]), target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesLambda(
            modes=np.array([[0, 0, 0], [1, 1, 1]]), target=np.array([[0, 1]])
        )


@pytest.mark.unit
def test_FixMode_False_or_None_modes():
    """Test Fixing Modes without specifying modes or All modes."""
    with pytest.raises(ValueError):
        FixModeR(modes=False, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixModeR(modes=None, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixModeZ(modes=False, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixModeZ(modes=None, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixModeLambda(modes=False, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixModeLambda(modes=None, target=np.array([[0, 1]]))


@pytest.mark.unit
def test_FixSumModes_False_or_None_modes():
    """Test Fixing Sum Modes without specifying modes or All modes."""
    with pytest.raises(ValueError):
        FixSumModesZ(modes=False, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesZ(modes=None, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesR(modes=False, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesR(modes=None, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesLambda(modes=False, target=np.array([[0, 1]]))
    with pytest.raises(ValueError):
        FixSumModesLambda(modes=None, target=np.array([[0, 1]]))


def _is_any_instance(things, cls):
    return any([isinstance(t, cls) for t in things])


@pytest.mark.unit
def test_FixAxis_util_correct_objectives():
    """Test util for fix axis constraints."""
    with pytest.warns(FutureWarning):
        cs = get_fixed_axis_constraints(iota=False)
    assert _is_any_instance(cs, FixAxisR)
    assert _is_any_instance(cs, FixAxisZ)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixPressure)
    assert _is_any_instance(cs, FixCurrent)

    with pytest.warns(FutureWarning):
        cs = get_fixed_axis_constraints(iota=True, kinetic=True)
    assert _is_any_instance(cs, FixAxisR)
    assert _is_any_instance(cs, FixAxisZ)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixElectronDensity)
    assert _is_any_instance(cs, FixElectronTemperature)
    assert _is_any_instance(cs, FixIonTemperature)
    assert _is_any_instance(cs, FixAtomicNumber)
    assert _is_any_instance(cs, FixIota)


@pytest.mark.unit
def test_FixNAE_util_correct_objectives():
    """Test util for fix NAE constraints."""
    eq = Equilibrium()
    qsc = Qsc.from_paper("precise QA")
    cs = get_NAE_constraints(eq, qsc, iota=False)
    assert _is_any_instance(cs, FixAxisR)
    assert _is_any_instance(cs, FixAxisZ)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixSumModesR)
    assert _is_any_instance(cs, FixSumModesZ)
    assert _is_any_instance(cs, FixPressure)
    assert _is_any_instance(cs, FixCurrent)

    cs = get_NAE_constraints(eq, qsc, iota=True, kinetic=True)
    assert _is_any_instance(cs, FixAxisR)
    assert _is_any_instance(cs, FixAxisZ)
    assert _is_any_instance(cs, FixPsi)
    assert _is_any_instance(cs, FixSumModesR)
    assert _is_any_instance(cs, FixSumModesZ)
    assert _is_any_instance(cs, FixElectronDensity)
    assert _is_any_instance(cs, FixElectronTemperature)
    assert _is_any_instance(cs, FixIonTemperature)
    assert _is_any_instance(cs, FixAtomicNumber)
    assert _is_any_instance(cs, FixIota)



"""Module for getting precomputed example equilibria."""

import os

import desc.io


def listall():
    """Return a list of examples that DESC has."""
    here = os.path.abspath(os.path.dirname(__file__))
    h5s = [f for f in os.listdir(here) if f.endswith(".h5")]
    names_stripped = [f.replace("_output.h5", "") for f in h5s]
    return names_stripped


def get(name, data=None):
    """Get example equilibria and data.

    Returns a solved equilibrium or selected attributes for one of several examples.

    A full list of valid names can be found with ``desc.examples.listall()``

    Parameters
    ----------
    name : str (case insensitive)
        Name of the example equilibrium to load from, should be one from
        ``desc.examples.listall()``.
    data : {None, "all", "boundary", "pressure", "iota", "current"}
        Data to return. None returns the final solved equilibrium. "all" returns the
        intermediate solutions from the continuation method as an EquilibriaFamily.
        "boundary" returns a representation for the last closed flux surface.
        "pressure", "iota", and "current" return the profile objects.

    Returns
    -------
    data : varies
        Data requested, see "data" argument for more details.

    """
    assert data in {None, "all", "boundary", "pressure", "iota", "current"}
    here = os.path.abspath(os.path.dirname(__file__))
    h5s = [f for f in os.listdir(here) if f.endswith(".h5")]
    h5s_lower = [f.lower() for f in h5s]
    filename = name.lower() + "_output.h5"
    try:
        idx = h5s_lower.index(filename)
    except ValueError as e:
        raise ValueError(
            "example {} not found, should be one of {}".format(name, listall())
        ) from e
    path = here + "/" + h5s[idx]
    assert os.path.exists(path)

    eqf = desc.io.load(path)

    if data is None:
        return eqf[-1]
    if data == "all":
        return eqf
    if data == "boundary":
        return eqf[-1].get_surface_at(rho=1)
    if data == "pressure":
        return eqf[-1].pressure
    if data == "iota":
        return eqf[-1].iota
    if data == "current":
        return eqf[-1].current



"""Functions for mapping between flux, sfl, and real space coordinates."""

import functools
import warnings

import numpy as np
from termcolor import colored

from desc.backend import fori_loop, jit, jnp, put, while_loop
from desc.compute import compute as compute_fun
from desc.compute import data_index, get_params, get_profiles, get_transforms
from desc.grid import ConcentricGrid, Grid, LinearGrid, QuadratureGrid
from desc.transform import Transform
from desc.utils import Index


def map_coordinates(  # noqa: C901
    eq,
    coords,
    inbasis,
    outbasis=("rho", "theta", "zeta"),
    guess=None,
    period=(np.inf, np.inf, np.inf),
    tol=1e-6,
    maxiter=30,
    **kwargs,
):
    """Given coordinates in inbasis, compute corresponding coordinates in outbasis.

    First solves for the computational coordinates that correspond to inbasis, then
    evaluates outbasis at those locations.

    Speed can often be significantly improved by providing a reasonable initial guess.
    The default is a nearest neighbor search on a coarse grid.

    NOTE: this function cannot be JIT compiled or differentiated with AD.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    coords : ndarray, shape(k,3)
        2D array of input coordinates. Each row is a different
        point in space.
    inbasis, outbasis : tuple of str
        Labels for input and output coordinates, eg ("R", "phi", "Z") or
        ("rho", "alpha", "zeta") or any combination thereof. Labels should be the
        same as the compute function data key
    guess : None or ndarray, shape(k,3)
        Initial guess for the computational coordinates ['rho', 'theta', 'zeta']
        corresponding to coords in inbasis. If None, heuristics are used based on
        in basis and a nearest neighbor search on a coarse grid.
    period : tuple of float
        Assumed periodicity for each quantity in inbasis.
        Use np.inf to denote no periodicity.
    tol : float
        Stopping tolerance.
    maxiter : int > 0
        Maximum number of Newton iterations

    Returns
    -------
    coords : ndarray, shape(k,3)
        Coordinates mapped from inbasis to outbasis. Values of NaN will be returned
        for coordinates where root finding did not succeed, possibly because the
        coordinate is not in the plasma volume.

    """
    inbasis = tuple(inbasis)
    outbasis = tuple(outbasis)
    assert (
        np.isfinite(maxiter) and maxiter > 0
    ), f"maxiter must be a positive integer, got {maxiter}"
    assert np.isfinite(tol) and tol > 0, f"tol must be a positive float, got {tol}"

    basis_derivs = tuple([f"{X}_{d}" for X in inbasis for d in ("r", "t", "z")])
    for key in basis_derivs:
        assert (
            key in data_index["desc.equilibrium.equilibrium.Equilibrium"]
        ), f"don't have recipe to compute partial derivative {key}"

    rhomin = kwargs.pop("rhomin", tol / 10)
    alpha = kwargs.pop("backtrack_frac", 0.1)
    maxls = kwargs.pop("max_backtrack", 10)
    assert len(kwargs) == 0, f"map_coordinates got unexpected kwargs: {kwargs.keys()}"
    period = np.asarray(period)
    coords = coords % period

    params = get_params(inbasis + basis_derivs, eq)

    @functools.partial(jit, static_argnums=1)
    def compute(y, basis):
        grid = Grid(y, sort=False, jitable=True)
        profiles = get_profiles(inbasis + basis_derivs, eq, grid, jitable=True)
        # do surface average to get iota once
        if "current" in profiles and profiles["current"] is not None:
            profiles["iota"] = eq.get_profile("iota")
        transforms = get_transforms(inbasis + basis_derivs, eq, grid, jitable=True)
        data = compute_fun(eq, basis, params, transforms, profiles)
        x = jnp.stack([data[k] for k in basis], axis=-1)
        return x

    @jit
    def residual(y):
        xk = compute(y, inbasis)
        r = xk % period - coords % period
        return jnp.where(r > period / 2, -period + r, r)

    @jit
    def jac(y):
        J = compute(y, basis_derivs)
        J = J.reshape((-1, 3, 3))
        return J

    @jit
    def fixup(y):
        r, t, z = y.T
        # negative rho -> flip theta
        t = jnp.where(r < 0, (t + np.pi), t)
        r = jnp.abs(r)
        r = jnp.clip(r, rhomin, 1)
        y = jnp.array([r, t, z]).T
        return y

    # use least squares to avoid singular behavior, cheap for 3x3 systems
    lstsq = jit(
        jnp.vectorize(
            lambda A, b: jnp.linalg.lstsq(A, b)[0], signature="(n,n),(n)->(n)"
        )
    )

    yk = guess
    if yk is None:
        # nearest neighbor search on coarse grid for initial guess
        yg = ConcentricGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP).nodes
        xg = compute(yg, inbasis)
        idx = jnp.zeros(len(coords)).astype(int)
        coords = jnp.asarray(coords)

        def _distance_body(i, idx):
            d = (coords[i] % period) - (xg % period)
            d = jnp.where(d > period / 2, period - d, d)
            distance = jnp.linalg.norm(d, axis=-1)
            k = jnp.argmin(distance)
            idx = put(idx, i, k)
            return idx

        idx = fori_loop(0, len(coords), _distance_body, idx)
        yk = yg[idx]

        # apply some heuristics based on common patterns
        if "rho" in inbasis:
            yk = put(yk.T, 0, coords[:, inbasis.index("rho")]).T
        if "theta" in inbasis:
            yk = put(yk.T, 1, coords[:, inbasis.index("theta")]).T
        elif "theta_PEST" in inbasis:  # lambda is usually small
            yk = put(yk.T, 1, coords[:, inbasis.index("theta_PEST")]).T
        if "zeta" in inbasis:
            yk = put(yk.T, 2, coords[:, inbasis.index("zeta")]).T
        elif "phi" in inbasis:
            yk = put(yk.T, 2, coords[:, inbasis.index("phi")]).T

    yk = fixup(yk)
    resk = residual(yk)

    for i in range(maxiter):
        if np.max(np.abs(resk)) < tol:
            break
        J = jac(yk)
        d = lstsq(J, resk)
        alphak = jnp.ones(yk.shape[0])
        for j in range(maxls):
            # backtracking line search
            yt = fixup(yk - alphak[:, None] * d)
            res = residual(yt)
            if np.max(np.abs(res)) < np.max(np.abs(resk)):
                yk = yt
                resk = res
                break
            else:
                # only reduce step size where needed
                alphak = jnp.where(
                    np.max(np.abs(res), axis=-1) > np.max(np.abs(resk), axis=-1),
                    alpha * alphak,
                    alphak,
                )
        else:
            # if we haven't found a good step, try taking a bad one and
            # hope the next one is better
            yk = yt
            resk = res

    r, t, z = yk.T
    res = jnp.max(jnp.abs(residual(yk)), axis=-1)
    r = jnp.where(res > tol, jnp.nan, r)
    t = jnp.where(res > tol, jnp.nan, t)
    z = jnp.where(res > tol, jnp.nan, z)

    yk = jnp.vstack([r, t, z]).T

    out = compute(yk, outbasis)

    return out


def compute_theta_coords(eq, flux_coords, L_lmn=None, tol=1e-6, maxiter=20):
    """Find geometric theta for given straight field line theta.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    flux_coords : ndarray, shape(k,3)
        2d array of flux coordinates [rho,theta*,zeta]. Each row is a different
        point in space.
    L_lmn : ndarray
        spectral coefficients for lambda. Defaults to eq.L_lmn
    tol : float
        Stopping tolerance.
    maxiter : int > 0
        maximum number of Newton iterations

    Returns
    -------
    coords : ndarray, shape(k,3)
        coordinates [rho,theta,zeta]. If Newton method doesn't converge for
        a given coordinate nan will be returned for those values

    """
    if L_lmn is None:
        L_lmn = eq.L_lmn
    rho, theta_star, zeta = flux_coords.T
    if maxiter <= 0:
        raise ValueError(f"maxiter must be a positive integer, got{maxiter}")
    if jnp.any(rho < 0):
        raise ValueError("rho values must be positive")

    # Note: theta* (also known as vartheta) is the poloidal straight field-line
    # angle in PEST-like flux coordinates

    nodes = flux_coords.copy()
    A0 = eq.L_basis.evaluate(nodes, (0, 0, 0))

    # theta* = theta + lambda
    lmbda = jnp.dot(A0, L_lmn)
    k = 0

    def cond_fun(nodes_k_lmbda):
        nodes, k, lmbda = nodes_k_lmbda
        theta_star_k = nodes[:, 1] + lmbda
        err = theta_star - theta_star_k
        return jnp.any(jnp.abs(err) > tol) & (k < maxiter)

    # Newton method for root finding
    def body_fun(nodes_k_lmbda):
        nodes, k, lmbda = nodes_k_lmbda
        A1 = eq.L_basis.evaluate(nodes, (0, 1, 0))
        lmbda_t = jnp.dot(A1, L_lmn)
        f = theta_star - nodes[:, 1] - lmbda
        df = -1 - lmbda_t
        nodes = put(nodes, Index[:, 1], nodes[:, 1] - f / df)
        A0 = eq.L_basis.evaluate(nodes, (0, 0, 0))
        lmbda = jnp.dot(A0, L_lmn)
        k += 1
        return (nodes, k, lmbda)

    nodes, k, lmbda = jit(while_loop, static_argnums=(0, 1))(
        cond_fun, body_fun, (nodes, k, lmbda)
    )
    theta_star_k = nodes[:, 1] + lmbda
    err = theta_star - theta_star_k
    noconverge = jnp.abs(err) > tol
    nodes = jnp.where(noconverge[:, np.newaxis], jnp.nan, nodes)

    return nodes


def compute_flux_coords(
    eq, real_coords, R_lmn=None, Z_lmn=None, tol=1e-6, maxiter=20, rhomin=1e-6
):
    """Find the (rho, theta, zeta) that correspond to given (R, phi, Z).

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    real_coords : ndarray, shape(k,3)
        2D array of real space coordinates [R,phi,Z]. Each row is a different
        point in space.
    R_lmn, Z_lmn : ndarray
        spectral coefficients for R and Z. Defaults to eq.R_lmn, eq.Z_lmn
    tol : float
        Stopping tolerance. Iterations stop when sqrt((R-Ri)**2 + (Z-Zi)**2) < tol
    maxiter : int > 0
        maximum number of Newton iterations
    rhomin : float
        minimum allowable value of rho (to avoid singularity at rho=0)

    Returns
    -------
    flux_coords : ndarray, shape(k,3)
        flux coordinates [rho,theta,zeta]. If Newton method doesn't converge for
        a given coordinate (often because it is outside the plasma boundary),
        nan will be returned for those values

    """
    if maxiter <= 0:
        raise ValueError(f"maxiter must be a positive integer, got{maxiter}")
    if R_lmn is None:
        R_lmn = eq.R_lmn
    if Z_lmn is None:
        Z_lmn = eq.Z_lmn

    R, phi, Z = real_coords.T
    R = jnp.abs(R)

    # nearest neighbor search on coarse grid for initial guess
    nodes = ConcentricGrid(L=20, M=10, N=0).nodes
    AR = eq.R_basis.evaluate(nodes)
    AZ = eq.Z_basis.evaluate(nodes)
    Rg = jnp.dot(AR, R_lmn)
    Zg = jnp.dot(AZ, Z_lmn)
    distance = (R[:, np.newaxis] - Rg) ** 2 + (Z[:, np.newaxis] - Zg) ** 2
    idx = jnp.argmin(distance, axis=1)

    rhok = nodes[idx, 0]
    thetak = nodes[idx, 1]
    Rk = Rg[idx]
    Zk = Zg[idx]
    k = 0

    def cond_fun(k_rhok_thetak_Rk_Zk):
        k, rhok, thetak, Rk, Zk = k_rhok_thetak_Rk_Zk
        return jnp.any(((R - Rk) ** 2 + (Z - Zk) ** 2) > tol**2) & (k < maxiter)

    def body_fun(k_rhok_thetak_Rk_Zk):
        k, rhok, thetak, Rk, Zk = k_rhok_thetak_Rk_Zk
        nodes = jnp.array([rhok, thetak, phi]).T
        ARr = eq.R_basis.evaluate(nodes, (1, 0, 0))
        Rr = jnp.dot(ARr, R_lmn)
        AZr = eq.Z_basis.evaluate(nodes, (1, 0, 0))
        Zr = jnp.dot(AZr, Z_lmn)
        ARt = eq.R_basis.evaluate(nodes, (0, 1, 0))
        Rt = jnp.dot(ARt, R_lmn)
        AZt = eq.Z_basis.evaluate(nodes, (0, 1, 0))
        Zt = jnp.dot(AZt, Z_lmn)

        tau = Rt * Zr - Rr * Zt
        eR = R - Rk
        eZ = Z - Zk
        thetak += (Zr * eR - Rr * eZ) / tau
        rhok += (Rt * eZ - Zt * eR) / tau
        # negative rho -> rotate theta instead
        thetak = jnp.where(
            rhok < 0, (thetak + np.pi) % (2 * np.pi), thetak % (2 * np.pi)
        )
        rhok = jnp.abs(rhok)
        rhok = jnp.clip(rhok, rhomin, 1)
        nodes = jnp.array([rhok, thetak, phi]).T

        AR = eq.R_basis.evaluate(nodes, (0, 0, 0))
        Rk = jnp.dot(AR, R_lmn)
        AZ = eq.Z_basis.evaluate(nodes, (0, 0, 0))
        Zk = jnp.dot(AZ, Z_lmn)
        k += 1
        return (k, rhok, thetak, Rk, Zk)

    k, rhok, thetak, Rk, Zk = while_loop(cond_fun, body_fun, (k, rhok, thetak, Rk, Zk))

    noconverge = (R - Rk) ** 2 + (Z - Zk) ** 2 > tol**2
    rho = jnp.where(noconverge, jnp.nan, rhok)
    theta = jnp.where(noconverge, jnp.nan, thetak)
    phi = jnp.where(noconverge, jnp.nan, phi)

    return jnp.vstack([rho, theta, phi]).T


def is_nested(eq, grid=None, R_lmn=None, Z_lmn=None, L_lmn=None, msg=None):
    """Check that an equilibrium has properly nested flux surfaces in a plane.

    Does so by checking coordinate Jacobian (sqrt(g)) sign.
    If coordinate Jacobian switches sign somewhere in the volume, this
    indicates that it is zero at some point, meaning surfaces are touching and
    the equilibrium is not nested.

    NOTE: If grid resolution used is too low, or the solution is just barely
    unnested, this function may fail to return the correct answer.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    grid  :  Grid, optional
        Grid on which to evaluate the coordinate Jacobian and check for the sign.
        (Default to QuadratureGrid with eq's current grid resolutions)
    R_lmn, Z_lmn, L_lmn : ndarray, optional
        spectral coefficients for R, Z, lambda. Defaults to eq.R_lmn, eq.Z_lmn
    msg : {None, "auto", "manual"}
        Warning to throw if unnested.

    Returns
    -------
    is_nested : bool
        whether the surfaces are nested

    """
    if R_lmn is None:
        R_lmn = eq.R_lmn
    if Z_lmn is None:
        Z_lmn = eq.Z_lmn
    if L_lmn is None:
        L_lmn = eq.L_lmn
    if grid is None:
        grid = QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP)

    transforms = get_transforms("sqrt(g)_PEST", obj=eq, grid=grid)
    data = compute_fun(
        "desc.equilibrium.equilibrium.Equilibrium",
        "sqrt(g)_PEST",
        params={
            "R_lmn": R_lmn,
            "Z_lmn": Z_lmn,
            "L_lmn": L_lmn,
        },
        transforms=transforms,
        profiles={},  # no profiles needed
    )

    nested = jnp.all(
        jnp.sign(data["sqrt(g)_PEST"][0]) == jnp.sign(data["sqrt(g)_PEST"])
    )
    if not nested:
        if msg == "auto":
            warnings.warn(
                colored(
                    "WARNING: Flux surfaces are no longer nested, exiting early. "
                    + "Automatic continuation method failed, consider specifying "
                    + "continuation steps manually",
                    "yellow",
                )
            )
        elif msg == "manual":
            warnings.warn(
                colored(
                    "WARNING: Flux surfaces are no longer nested, exiting early."
                    + "Consider taking smaller perturbation/resolution steps "
                    + "or reducing trust radius",
                    "yellow",
                )
            )
    return nested


def to_sfl(
    eq,
    L=None,
    M=None,
    N=None,
    L_grid=None,
    M_grid=None,
    N_grid=None,
    rcond=None,
    copy=False,
):
    """Transform this equilibrium to use straight field line coordinates.

    Uses a least squares fit to find FourierZernike coefficients of R, Z, Rb, Zb
    with respect to the straight field line coordinates, rather than the boundary
    coordinates. The new lambda value will be zero.

    NOTE: Though the converted equilibrium will have the same flux surfaces,
    the force balance error will likely be higher than the original equilibrium.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to use
    L : int, optional
        radial resolution to use for SFL equilibrium. Default = 1.5*eq.L
    M : int, optional
        poloidal resolution to use for SFL equilibrium. Default = 1.5*eq.M
    N : int, optional
        toroidal resolution to use for SFL equilibrium. Default = 1.5*eq.N
    L_grid : int, optional
        radial spatial resolution to use for fit to new basis. Default = 2*L
    M_grid : int, optional
        poloidal spatial resolution to use for fit to new basis. Default = 2*M
    N_grid : int, optional
        toroidal spatial resolution to use for fit to new basis. Default = 2*N
    rcond : float, optional
        cutoff for small singular values in the least squares fit.
    copy : bool, optional
        Whether to update the existing equilibrium or make a copy (Default).

    Returns
    -------
    eq_sfl : Equilibrium
        Equilibrium transformed to a straight field line coordinate representation.

    """
    L = L or int(1.5 * eq.L)
    M = M or int(1.5 * eq.M)
    N = N or int(1.5 * eq.N)
    L_grid = L_grid or int(2 * L)
    M_grid = M_grid or int(2 * M)
    N_grid = N_grid or int(2 * N)

    grid = ConcentricGrid(L_grid, M_grid, N_grid, node_pattern="ocs")
    bdry_grid = LinearGrid(M=M, N=N, rho=1.0)

    toroidal_coords = eq.compute(["R", "Z", "lambda"], grid=grid)
    theta = grid.nodes[:, 1]
    vartheta = theta + toroidal_coords["lambda"]
    sfl_grid = grid
    sfl_grid.nodes[:, 1] = vartheta

    bdry_coords = eq.compute(["R", "Z", "lambda"], grid=bdry_grid)
    bdry_theta = bdry_grid.nodes[:, 1]
    bdry_vartheta = bdry_theta + bdry_coords["lambda"]
    bdry_sfl_grid = bdry_grid
    bdry_sfl_grid.nodes[:, 1] = bdry_vartheta

    if copy:
        eq_sfl = eq.copy()
    else:
        eq_sfl = eq
    eq_sfl.change_resolution(L, M, N)

    R_sfl_transform = Transform(
        sfl_grid, eq_sfl.R_basis, build=False, build_pinv=True, rcond=rcond
    )
    R_lmn_sfl = R_sfl_transform.fit(toroidal_coords["R"])
    del R_sfl_transform  # these can take up a lot of memory so delete when done.

    Z_sfl_transform = Transform(
        sfl_grid, eq_sfl.Z_basis, build=False, build_pinv=True, rcond=rcond
    )
    Z_lmn_sfl = Z_sfl_transform.fit(toroidal_coords["Z"])
    del Z_sfl_transform
    L_lmn_sfl = np.zeros_like(eq_sfl.L_lmn)

    R_sfl_bdry_transform = Transform(
        bdry_sfl_grid,
        eq_sfl.surface.R_basis,
        build=False,
        build_pinv=True,
        rcond=rcond,
    )
    Rb_lmn_sfl = R_sfl_bdry_transform.fit(bdry_coords["R"])
    del R_sfl_bdry_transform

    Z_sfl_bdry_transform = Transform(
        bdry_sfl_grid,
        eq_sfl.surface.Z_basis,
        build=False,
        build_pinv=True,
        rcond=rcond,
    )
    Zb_lmn_sfl = Z_sfl_bdry_transform.fit(bdry_coords["Z"])
    del Z_sfl_bdry_transform

    eq_sfl.Rb_lmn = Rb_lmn_sfl
    eq_sfl.Zb_lmn = Zb_lmn_sfl
    eq_sfl.R_lmn = R_lmn_sfl
    eq_sfl.Z_lmn = Z_lmn_sfl
    eq_sfl.L_lmn = L_lmn_sfl

    return eq_sfl



"""Utilities for getting standard groups of objectives and constraints."""

from ._equilibrium import (
    CurrentDensity,
    Energy,
    ForceBalance,
    HelicalForceBalance,
    RadialForceBalance,
)
from .linear_objectives import (
    AxisRSelfConsistency,
    AxisZSelfConsistency,
    BoundaryRSelfConsistency,
    BoundaryZSelfConsistency,
    FixAnisotropy,
    FixAtomicNumber,
    FixAxisR,
    FixAxisZ,
    FixBoundaryR,
    FixBoundaryZ,
    FixCurrent,
    FixElectronDensity,
    FixElectronTemperature,
    FixIonTemperature,
    FixIota,
    FixLambdaGauge,
    FixPressure,
    FixPsi,
)
from .nae_utils import calc_zeroth_order_lambda, make_RZ_cons_1st_order
from .objective_funs import ObjectiveFunction


def get_equilibrium_objective(eq=None, mode="force", normalize=True):
    """Get the objective function for a typical force balance equilibrium problem.

    Parameters
    ----------
    mode : one of {"force", "forces", "energy", "vacuum"}
        which objective to return. "force" computes force residuals on unified grid.
        "forces" uses two different grids for radial and helical forces. "energy" is
        for minimizing MHD energy. "vacuum" directly minimizes current density.
    normalize : bool
        Whether to normalize units of objective.

    Returns
    -------
    objective, ObjectiveFunction
        An objective function with default force balance objectives.
    """
    if mode == "energy":
        objectives = Energy(eq=eq, normalize=normalize, normalize_target=normalize)
    elif mode == "force":
        objectives = ForceBalance(
            eq=eq, normalize=normalize, normalize_target=normalize
        )
    elif mode == "forces":
        objectives = (
            RadialForceBalance(eq=eq, normalize=normalize, normalize_target=normalize),
            HelicalForceBalance(eq=eq, normalize=normalize, normalize_target=normalize),
        )
    elif mode == "vacuum":
        objectives = CurrentDensity(
            eq=eq, normalize=normalize, normalize_target=normalize
        )
    else:
        raise ValueError("got an unknown equilibrium objective type '{}'".format(mode))
    return ObjectiveFunction(objectives)


def get_fixed_axis_constraints(
    eq=None, profiles=True, iota=True, kinetic=False, anisotropy=False, normalize=True
):
    """Get the constraints necessary for a fixed-axis equilibrium problem.

    Parameters
    ----------
    profiles : bool
        Whether to also return constraints to fix input profiles.
    iota : bool
        Whether to add FixIota or FixCurrent as a constraint.
    kinetic : bool
        Whether to add constraints to fix kinetic profiles or pressure
    anisotropy : bool
        Whether to add constraint to fix anisotropic pressure.
    normalize : bool
        Whether to apply constraints in normalized units.

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-axis problems.

    """
    constraints = (
        FixAxisR(eq=eq, normalize=normalize, normalize_target=normalize),
        FixAxisZ(eq=eq, normalize=normalize, normalize_target=normalize),
        FixPsi(eq=eq, normalize=normalize, normalize_target=normalize),
    )
    if profiles:
        if kinetic:
            constraints += (
                FixElectronDensity(
                    eq=eq, normalize=normalize, normalize_target=normalize
                ),
                FixElectronTemperature(
                    eq=eq, normalize=normalize, normalize_target=normalize
                ),
                FixIonTemperature(
                    eq=eq, normalize=normalize, normalize_target=normalize
                ),
                FixAtomicNumber(eq=eq, normalize=normalize, normalize_target=normalize),
            )
        else:
            constraints += (
                FixPressure(eq=eq, normalize=normalize, normalize_target=normalize),
            )
            if anisotropy:
                constraints += (
                    FixAnisotropy(
                        eq=eq, normalize=normalize, normalize_target=normalize
                    ),
                )

        if iota:
            constraints += (
                FixIota(eq=eq, normalize=normalize, normalize_target=normalize),
            )
        else:
            constraints += (
                FixCurrent(eq=eq, normalize=normalize, normalize_target=normalize),
            )
    return constraints


def get_fixed_boundary_constraints(
    eq=None, profiles=True, iota=True, kinetic=False, anisotropy=False, normalize=True
):
    """Get the constraints necessary for a typical fixed-boundary equilibrium problem.

    Parameters
    ----------
    eq : Equilibrium
        Equilibrium to constraint.
    profiles : bool
        Whether to also return constraints to fix input profiles.
    iota : bool
        Whether to add FixIota or FixCurrent as a constraint.
    kinetic : bool
        Whether to also fix kinetic profiles.
    anisotropy : bool
        Whether to add constraint to fix anisotropic pressure.
    normalize : bool
        Whether to apply constraints in normalized units.

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-boundary problems.

    """
    constraints = (
        FixBoundaryR(eq=eq, normalize=normalize, normalize_target=normalize),
        FixBoundaryZ(eq=eq, normalize=normalize, normalize_target=normalize),
        FixPsi(eq=eq, normalize=normalize, normalize_target=normalize),
    )
    if profiles:
        if kinetic:
            constraints += (
                FixElectronDensity(
                    eq=eq, normalize=normalize, normalize_target=normalize
                ),
                FixElectronTemperature(
                    eq=eq, normalize=normalize, normalize_target=normalize
                ),
                FixIonTemperature(
                    eq=eq, normalize=normalize, normalize_target=normalize
                ),
                FixAtomicNumber(eq=eq, normalize=normalize, normalize_target=normalize),
            )
        else:
            constraints += (
                FixPressure(eq=eq, normalize=normalize, normalize_target=normalize),
            )
            if anisotropy:
                constraints += (
                    FixAnisotropy(
                        eq=eq, normalize=normalize, normalize_target=normalize
                    ),
                )
        if iota:
            constraints += (
                FixIota(eq=eq, normalize=normalize, normalize_target=normalize),
            )
        else:
            constraints += (
                FixCurrent(eq=eq, normalize=normalize, normalize_target=normalize),
            )
    return constraints


def get_NAE_constraints(
    desc_eq,
    qsc_eq,
    order=1,
    profiles=True,
    iota=False,
    kinetic=False,
    anisotropy=False,
    normalize=True,
    N=None,
    fix_lambda=False,
):
    """Get the constraints necessary for fixing NAE behavior in an equilibrium problem.

    Parameters
    ----------
    desc_eq : Equilibrium
        Equilibrium to constrain behavior of
        (assumed to be a fit from the NAE equil using `.from_near_axis()`).
    qsc_eq : Qsc
        Qsc object defining the near-axis equilibrium to constrain behavior to.
    order : int
        order (in rho) of near-axis behavior to constrain
    profiles : bool
        Whether to also return constraints to fix input profiles.
    iota : bool
        Whether to add `FixIota` or `FixCurrent` as a constraint.
    kinetic : bool
        Whether to also fix kinetic profiles.
    anisotropy : bool
        Whether to add constraint to fix anisotropic pressure.
    normalize : bool
        Whether to apply constraints in normalized units.
    N : int
        max toroidal resolution to constrain.
        If `None`, defaults to equilibrium's toroidal resolution
    fix_lambda : bool or int
        Whether to constrain lambda to match that of the NAE near-axis
        if an `int`, fixes lambda up to that order in rho {0,1}
        if `True`, fixes lambda up to the specified order given by `order`

    Returns
    -------
    constraints, tuple of _Objectives
        A list of the linear constraints used in fixed-axis problems.
    """
    if not isinstance(fix_lambda, bool):
        fix_lambda = int(fix_lambda)
    constraints = (
        FixAxisR(eq=desc_eq, normalize=normalize, normalize_target=normalize),
        FixAxisZ(eq=desc_eq, normalize=normalize, normalize_target=normalize),
        FixPsi(eq=desc_eq, normalize=normalize, normalize_target=normalize),
    )
    if profiles:
        if kinetic:
            constraints += (
                FixElectronDensity(
                    eq=desc_eq, normalize=normalize, normalize_target=normalize
                ),
                FixElectronTemperature(
                    eq=desc_eq, normalize=normalize, normalize_target=normalize
                ),
                FixIonTemperature(
                    eq=desc_eq, normalize=normalize, normalize_target=normalize
                ),
                FixAtomicNumber(
                    eq=desc_eq, normalize=normalize, normalize_target=normalize
                ),
            )
        else:
            constraints += (
                FixPressure(
                    eq=desc_eq, normalize=normalize, normalize_target=normalize
                ),
            )
            if anisotropy:
                constraints += (
                    FixAnisotropy(
                        eq=desc_eq, normalize=normalize, normalize_target=normalize
                    ),
                )

        if iota:
            constraints += (
                FixIota(eq=desc_eq, normalize=normalize, normalize_target=normalize),
            )
        else:
            constraints += (
                FixCurrent(eq=desc_eq, normalize=normalize, normalize_target=normalize),
            )
    if fix_lambda or (fix_lambda >= 0 and type(fix_lambda) is int):
        L_axis_constraints, _, _ = calc_zeroth_order_lambda(
            qsc=qsc_eq, desc_eq=desc_eq, N=N
        )
        constraints += L_axis_constraints
    if order >= 1:  # first order constraints
        constraints += make_RZ_cons_1st_order(
            qsc=qsc_eq, desc_eq=desc_eq, N=N, fix_lambda=fix_lambda and fix_lambda > 0
        )
    if order >= 2:  # 2nd order constraints
        raise NotImplementedError("NAE constraints only implemented up to O(rho) ")

    return constraints


def maybe_add_self_consistency(eq, constraints):
    """Add self consistency constraints if needed."""

    def _is_any_instance(things, cls):
        return any([isinstance(t, cls) for t in things])

    if not _is_any_instance(constraints, BoundaryRSelfConsistency):
        constraints += (BoundaryRSelfConsistency(eq=eq),)
    if not _is_any_instance(constraints, BoundaryZSelfConsistency):
        constraints += (BoundaryZSelfConsistency(eq=eq),)
    if not _is_any_instance(constraints, FixLambdaGauge):
        constraints += (FixLambdaGauge(eq=eq),)
    if not _is_any_instance(constraints, AxisRSelfConsistency):
        constraints += (AxisRSelfConsistency(eq=eq),)
    if not _is_any_instance(constraints, AxisZSelfConsistency):
        constraints += (AxisZSelfConsistency(eq=eq),)
    return constraints



"""Legacy functions for reading and writing ascii format."""

import numpy as np


def write_ascii(fname, eq):
    """Print the equilibrium solution to a text file.

    Parameters
    ----------
    fname : str or path-like
        filename of output file.
    eq : dict
        dictionary of equilibrium parameters.

    """
    # can't import other stuff in io due to circular imports, so have to check by name
    assert (
        eq.pressure.__class__.__name__ == "PowerSeriesProfile"
        and eq.iota.__class__.__name__ == "PowerSeriesProfile"
    ), "Equilibrium must have power series profiles for ascii io"

    pressure = eq.pressure
    iota = eq.iota
    L = max(pressure.basis.L, iota.basis.L)
    pressure.change_resolution(L)
    iota.change_resolution(L)

    # open file
    file = open(fname, "w+")
    file.seek(0)

    # scaling factors
    file.write("NFP = {:3d}\n".format(int(eq.NFP)))
    file.write("Psi = {:16.8E}\n".format(eq.Psi))

    # boundary parameters
    if eq.sym:
        nbdry = len(np.nonzero(eq.Rb_lmn)[0]) + len(np.nonzero(eq.Zb_lmn)[0])
        file.write("Nbdry = {:3d}\n".format(nbdry))
        for k, (l, m, n) in enumerate(eq.surface.R_basis.modes):
            if eq.Rb_lmn[k] != 0:
                file.write(
                    "m: {:3d} n: {:3d} bR = {:16.8E} bZ = {:16.8E}\n".format(
                        m, n, eq.Rb_lmn[k], 0
                    )
                )
        for k, (l, m, n) in enumerate(eq.surface.Z_basis.modes):
            if eq.Zb_lmn[k] != 0:
                file.write(
                    "m: {:3d} n: {:3d} bR = {:16.8E} bZ = {:16.8E}\n".format(
                        m, n, 0, eq.Zb_lmn[k]
                    )
                )
    else:
        nbdry = eq.surface.R_basis.num_modes
        file.write("Nbdry = {:3d}\n".format(nbdry))
        for k, (l, m, n) in enumerate(eq.surface.R_basis.modes):
            file.write(
                "m: {:3d} n: {:3d} bR = {:16.8E} bZ = {:16.8E}\n".format(
                    m, n, eq.Rb_lmn[k], eq.Zb_lmn[k]
                )
            )

    # profile coefficients
    nprof = len(np.nonzero(np.abs(pressure.params) + np.abs(iota.params))[0])
    file.write("Nprof = {:3d}\n".format(nprof))
    for k, (l, m, n) in enumerate(pressure.basis.modes):
        if (pressure.params[k] != 0) or (iota.params[k] != 0):
            file.write(
                "l: {:3d} cP = {:16.8E} cI = {:16.8E}\n".format(
                    k, pressure.params[k], iota.params[k]
                )
            )

    # R, Z & L Fourier-Zernike coefficients
    if eq.sym:
        nRZ = eq.R_basis.num_modes + eq.Z_basis.num_modes
        file.write("NRZ = {:5d}\n".format(nRZ))
        for k, (l, m, n) in enumerate(eq.R_basis.modes):
            file.write(
                "l: {:3d} m: {:3d} n: {:3d} ".format(l, m, n)
                + "cR = {:16.8E} cZ = {:16.8E} cL = {:16.8E}\n".format(
                    eq.R_lmn[k], 0, 0
                )
            )
        for k, (l, m, n) in enumerate(eq.Z_basis.modes):
            file.write(
                "l: {:3d} m: {:3d} n: {:3d} ".format(l, m, n)
                + "cR = {:16.8E} cZ = {:16.8E} cL = {:16.8E}\n".format(
                    0, eq.Z_lmn[k], eq.L_lmn[k]
                )
            )
    else:
        nRZ = eq.R_basis.num_modes
        file.write("NRZ = {:5d}\n".format(nRZ))
        for k, (l, m, n) in enumerate(eq.R_basis.modes):
            file.write(
                "l: {:3d} m: {:3d} n: {:3d} ".format(l, m, n)
                + "cR = {:16.8E} cZ = {:16.8E} cL = {:16.8E}\n".format(
                    eq.R_lmn[k], eq.Z_lmn[k], eq.L_lmn[k]
                )
            )

    # close file
    file.truncate()
    file.close()


def read_ascii(filename):
    """Read a previously generated DESC ascii output file.

    Parameters
    ----------
    filename : str or path-like
        path to file to read

    Returns
    -------
    eq : dict
        dictionary of equilibrium parameters.

    """
    from desc.backend import sign
    from desc.equilibrium import Equilibrium
    from desc.utils import copy_coeffs

    eq = {}
    f = open(filename)
    lines = list(f)
    eq["NFP"] = int(lines[0].strip("\n").split()[-1])
    eq["Psi"] = float(lines[1].strip("\n").split()[-1])
    lines = lines[2:]

    Nbdry = int(lines[0].strip("\n").split()[-1])
    bdry_idx = np.zeros((Nbdry, 2), dtype=int)
    bdryR = np.zeros(Nbdry)
    bdryZ = np.zeros(Nbdry)
    for i in range(Nbdry):
        bdry_idx[i, 0] = int(lines[i + 1].strip("\n").split()[1])
        bdry_idx[i, 1] = int(lines[i + 1].strip("\n").split()[3])
        bdryR[i] = float(lines[i + 1].strip("\n").split()[6])
        bdryZ[i] = float(lines[i + 1].strip("\n").split()[9])
    eq["surface"] = np.hstack(
        [
            np.zeros((Nbdry, 1)),
            bdry_idx,
            bdryR.reshape((-1, 1)),
            bdryZ.reshape((-1, 1)),
        ]
    )
    lines = lines[Nbdry + 1 :]

    Nprof = int(lines[0].strip("\n").split()[-1])
    pl = np.zeros(Nprof).astype(int)
    cP = np.zeros(Nprof)
    cI = np.zeros(Nprof)
    for i in range(Nprof):
        pl[i] = int(lines[i + 1].strip("\n").split()[1])
        cP[i] = float(lines[i + 1].strip("\n").split()[4])
        cI[i] = float(lines[i + 1].strip("\n").split()[7])
    eq["pressure"] = np.hstack([pl.reshape((-1, 1)), cP.reshape((-1, 1))])
    eq["iota"] = np.hstack([pl.reshape((-1, 1)), cI.reshape((-1, 1))])
    lines = lines[Nprof + 1 :]

    NRZ = int(lines[0].strip("\n").split()[-1])
    zern_idx = np.zeros((NRZ, 3), dtype=int)
    cR = np.zeros(NRZ)
    cZ = np.zeros(NRZ)
    cL = np.zeros(NRZ)
    for i in range(NRZ):
        zern_idx[i, 0] = int(lines[i + 1].strip("\n").split()[1])
        zern_idx[i, 1] = int(lines[i + 1].strip("\n").split()[3])
        zern_idx[i, 2] = int(lines[i + 1].strip("\n").split()[5])
        cR[i] = float(lines[i + 1].strip("\n").split()[8])
        cZ[i] = float(lines[i + 1].strip("\n").split()[11])
        cL[i] = float(lines[i + 1].strip("\n").split()[14])
    lines = lines[NRZ + 1 :]

    eq["L"] = np.max(abs(zern_idx[:, 0]))
    eq["M"] = np.max(abs(zern_idx[:, 1]))
    eq["N"] = np.max(abs(zern_idx[:, 2]))

    if np.all(
        cR[np.where(sign(zern_idx[:, 1]) != sign(zern_idx[:, 2]))] == 0
    ) and np.all(cZ[np.where(sign(zern_idx[:, 1]) == sign(zern_idx[:, 2]))] == 0):
        eq["sym"] = True
    else:
        eq["sym"] = False

    eq["spectral_indexing"] = "ansi" if eq["L"] == eq["M"] else "fringe"
    equil = Equilibrium(**eq)
    equil.R_lmn = copy_coeffs(cR, zern_idx, equil.R_basis.modes)
    equil.Z_lmn = copy_coeffs(cZ, zern_idx, equil.Z_basis.modes)
    equil.L_lmn = copy_coeffs(cL, zern_idx, equil.L_basis.modes)

    return equil



"""Functions and classes for reading and writing DESC data."""

# InputReader lives outside this module for import ordering reasons, so we can
# import InputReader in __main__ without importing equilibrium_io which imports JAX
# stuff potentially before we've set the GPU correctly.
# We include a link to it here for backwards compatibility
from desc.input_reader import InputReader

from .ascii_io import read_ascii, write_ascii
from .equilibrium_io import IOAble, load
from .hdf5_io import hdf5Reader, hdf5Writer
from .pickle_io import PickleReader, PickleWriter

__all__ = ["InputReader", "load"]



"""Generic objectives that don't belong anywhere else."""

import inspect
import re

from desc.backend import jnp
from desc.compute import compute as compute_fun
from desc.compute import data_index
from desc.compute.utils import get_params, get_profiles, get_transforms
from desc.grid import QuadratureGrid

from .objective_funs import _Objective


class ObjectiveFromUser(_Objective):
    """Wrap a user defined objective function.

    The user supplied function should take two arguments: ``grid`` and ``data``.

    ``grid`` is the Grid object containing the nodes where the data is computed.

    ``data`` is a dictionary of values with keys from the list of `variables`_. Values
    will be the given data evaluated at ``grid``.

    The function should be JAX traceable and differentiable, and should return a single
    JAX array. The source code of the function must be visible to the ``inspect`` module
    for parsing.

    .. _variables: https://desc-docs.readthedocs.io/en/stable/variables.html

    Parameters
    ----------
    fun : callable
        Custom objective function.
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
        Has no effect for this objective.
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True. Has no effect for this objective.
    grid : Grid, optional
        Collocation grid containing the nodes to evaluate at.
    name : str, optional
        Name of the objective function.


    Examples
    --------
    .. code-block:: python

        from desc.compute.utils import surface_averages
        def myfun(grid, data):
            # This will compute the flux surface average of the function
            # R*B_T from the Grad-Shafranov equation
            f = data['R']*data['B_phi']
            f_fsa = surface_averages(grid, f, sqrt_g=data['sqrt_g'])
            # this has the FSA values on the full grid, but we just want
            # the unique values:
            return grid.compress(f_fsa)

        myobj = ObjectiveFromUser(myfun)

    """

    _units = "(Unknown)"
    _print_value_fmt = "Custom Objective value: {:10.3e} "

    def __init__(
        self,
        fun,
        eq=None,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        grid=None,
        name="custom",
    ):
        if target is None and bounds is None:
            target = 0
        self._fun = fun
        self._grid = grid
        super().__init__(
            eq=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )

    def build(self, eq=None, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        eq : Equilibrium, optional
            Equilibrium that will be optimized to satisfy the Objective.
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = eq or self._eq
        if self._grid is None:
            grid = QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP)
        else:
            grid = self._grid

        def get_vars(fun):
            pattern = r"data\[(.*?)\]"
            src = inspect.getsource(fun)
            variables = re.findall(pattern, src)
            variables = list({s.strip().strip("'").strip('"') for s in variables})
            return variables

        self._data_keys = get_vars(self._fun)
        dummy_data = {}
        p = "desc.equilibrium.equilibrium.Equilibrium"
        for key in self._data_keys:
            assert key in data_index[p], f"Don't know how to compute {key}"
            if data_index[p][key]["dim"] == 0:
                dummy_data[key] = jnp.array(0.0)
            else:
                dummy_data[key] = jnp.empty((grid.num_nodes, data_index[p][key]["dim"]))

        self._fun_wrapped = lambda data: self._fun(grid, data)
        import jax

        self._dim_f = jax.eval_shape(self._fun_wrapped, dummy_data).size
        self._scalar = self._dim_f == 1
        self._args = get_params(
            self._data_keys,
            obj="desc.equilibrium.equilibrium.Equilibrium",
            has_axis=grid.axis.size,
        )
        profiles = get_profiles(self._data_keys, obj=eq, grid=grid)
        transforms = get_transforms(self._data_keys, obj=eq, grid=grid)
        self._constants = {
            "transforms": transforms,
            "profiles": profiles,
        }

        super().build(eq=eq, use_jit=use_jit, verbose=verbose)

    def compute(self, *args, **kwargs):
        """Compute the quantity.

        Parameters
        ----------
        args : ndarray
            Parameters given by self.args.

        Returns
        -------
        f : ndarray
            Computed quantity.

        """
        params, constants = self._parse_args(*args, **kwargs)
        if constants is None:
            constants = self.constants
        data = compute_fun(
            "desc.equilibrium.equilibrium.Equilibrium",
            self._data_keys,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
        )
        f = self._fun_wrapped(data)
        return f


class GenericObjective(_Objective):
    """A generic objective that can compute any quantity from the `data_index`.

    Parameters
    ----------
    f : str
        Name of the quantity to compute.
    eq : Equilibrium
        Equilibrium that will be optimized to satisfy the Objective.
    target : {float, ndarray}, optional
        Target value(s) of the objective. Only used if bounds is None.
        Must be broadcastable to Objective.dim_f.
    bounds : tuple of {float, ndarray}, optional
        Lower and upper bounds on the objective. Overrides target.
        Both bounds must be broadcastable to to Objective.dim_f
    weight : {float, ndarray}, optional
        Weighting to apply to the Objective, relative to other Objectives.
        Must be broadcastable to to Objective.dim_f
    normalize : bool, optional
        Whether to compute the error in physical units or non-dimensionalize.
        Has no effect for this objective
    normalize_target : bool, optional
        Whether target and bounds should be normalized before comparing to computed
        values. If `normalize` is `True` and the target is in physical units,
        this should also be set to True. Has no effect for this objective.
    grid : Grid, optional
        Collocation grid containing the nodes to evaluate at.
    name : str, optional
        Name of the objective function.

    """

    _print_value_fmt = "GenericObjective value: {:10.3e} "

    def __init__(
        self,
        f,
        eq=None,
        target=None,
        bounds=None,
        weight=1,
        normalize=True,
        normalize_target=True,
        grid=None,
        name="generic",
    ):
        if target is None and bounds is None:
            target = 0
        self.f = f
        self._grid = grid
        super().__init__(
            eq=eq,
            target=target,
            bounds=bounds,
            weight=weight,
            normalize=normalize,
            normalize_target=normalize_target,
            name=name,
        )
        self._scalar = not bool(
            data_index["desc.equilibrium.equilibrium.Equilibrium"][self.f]["dim"]
        )
        self._coordinates = data_index["desc.equilibrium.equilibrium.Equilibrium"][
            self.f
        ]["coordinates"]
        self._units = (
            "("
            + data_index["desc.equilibrium.equilibrium.Equilibrium"][self.f]["units"]
            + ")"
        )

    def build(self, eq=None, use_jit=True, verbose=1):
        """Build constant arrays.

        Parameters
        ----------
        eq : Equilibrium, optional
            Equilibrium that will be optimized to satisfy the Objective.
        use_jit : bool, optional
            Whether to just-in-time compile the objective and derivatives.
        verbose : int, optional
            Level of output.

        """
        eq = eq or self._eq
        if self._grid is None:
            grid = QuadratureGrid(eq.L_grid, eq.M_grid, eq.N_grid, eq.NFP)
        else:
            grid = self._grid

        p = "desc.equilibrium.equilibrium.Equilibrium"
        if data_index[p][self.f]["dim"] == 0:
            self._dim_f = 1
        elif data_index[p][self.f]["coordinates"] == "r":
            self._dim_f = grid.num_rho
        else:
            self._dim_f = grid.num_nodes * data_index[p][self.f]["dim"]
        self._args = get_params(
            self.f,
            obj="desc.equilibrium.equilibrium.Equilibrium",
            has_axis=grid.axis.size,
        )
        profiles = get_profiles(self.f, obj=eq, grid=grid)
        transforms = get_transforms(self.f, obj=eq, grid=grid)
        self._constants = {
            "transforms": transforms,
            "profiles": profiles,
        }

        super().build(eq=eq, use_jit=use_jit, verbose=verbose)

    def compute(self, *args, **kwargs):
        """Compute the quantity.

        Parameters
        ----------
        args : ndarray
            Parameters given by self.args.

        Returns
        -------
        f : ndarray
            Computed quantity.

        """
        params, constants = self._parse_args(*args, **kwargs)
        if constants is None:
            constants = self.constants
        data = compute_fun(
            "desc.equilibrium.equilibrium.Equilibrium",
            self.f,
            params=params,
            transforms=constants["transforms"],
            profiles=constants["profiles"],
        )
        f = data[self.f]
        if self._coordinates == "r":
            f = constants["transforms"]["grid"].compress(f, surface_label="rho")
        return f.flatten(order="F")  # so that default quad weights line up correctly.



"""Tests for different Curve classes."""

import numpy as np
import pytest

from desc.equilibrium import Equilibrium
from desc.geometry import (
    FourierPlanarCurve,
    FourierRZCurve,
    FourierXYZCurve,
    SplineXYZCurve,
)
from desc.grid import Grid, LinearGrid
from desc.io import InputReader


class TestRZCurve:
    """Tests for FourierRZCurve class."""

    @pytest.mark.unit
    def test_length(self):
        """Test length of circular curve."""
        c = FourierRZCurve()
        np.testing.assert_allclose(
            c.compute("length", grid=20)["length"], 10 * 2 * np.pi
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("length", grid=20)["length"], 10 * 2 * np.pi
        )

    @pytest.mark.unit
    def test_curvature(self):
        """Test curvature of circular curve."""
        c = FourierRZCurve()
        np.testing.assert_allclose(c.compute("curvature", grid=20)["curvature"], 1 / 10)
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(c.compute("curvature", grid=20)["curvature"], 1 / 10)

    @pytest.mark.unit
    def test_torsion(self):
        """Test torsion of circular curve."""
        c = FourierRZCurve()
        np.testing.assert_allclose(c.compute("torsion", grid=20)["torsion"], 0)
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(c.compute("torsion", grid=20)["torsion"], 0)

    @pytest.mark.unit
    def test_frenet(self):
        """Test frenet-serret frame of circular curve."""
        c = FourierRZCurve()
        data = c.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"], basis="rpz", grid=0
        )
        T, N, B = data["frenet_tangent"], data["frenet_normal"], data["frenet_binormal"]
        np.testing.assert_allclose(T, np.array([[0, 1, 0]]), atol=1e-12)
        np.testing.assert_allclose(N, np.array([[-1, 0, 0]]), atol=1e-12)
        np.testing.assert_allclose(B, np.array([[0, 0, 1]]), atol=1e-12)
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        data = c.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"], basis="xyz", grid=0
        )
        T, N, B = data["frenet_tangent"], data["frenet_normal"], data["frenet_binormal"]
        np.testing.assert_allclose(T, np.array([[0, 1, 0]]), atol=1e-12)
        np.testing.assert_allclose(N, np.array([[1, 0, 0]]), atol=1e-12)
        np.testing.assert_allclose(B, np.array([[0, 0, 1]]), atol=1e-12)

    @pytest.mark.unit
    def test_coords(self):
        """Test lab frame coordinates of circular curve."""
        c = FourierRZCurve()
        x, y, z = c.compute("x", grid=0, basis="xyz")["x"].T
        np.testing.assert_allclose(x, 10)
        np.testing.assert_allclose(y, 0)
        np.testing.assert_allclose(z, 0)
        c.rotate(angle=np.pi / 2)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        r, p, z = c.compute("x", grid=0, basis="rpz")["x"].T
        np.testing.assert_allclose(r, np.sqrt(1**2 + 9**2))
        np.testing.assert_allclose(p, np.arctan2(-9, 1))
        np.testing.assert_allclose(z, 1)

    @pytest.mark.unit
    def test_misc(self):
        """Test getting/setting misc attributes of FourierRZCurve."""
        c = FourierRZCurve()

        R, Z = c.get_coeffs(0)
        np.testing.assert_allclose(R, 10)
        np.testing.assert_allclose(Z, 0)
        c.set_coeffs(0, 5, None)
        np.testing.assert_allclose(c.R_n, [5])
        np.testing.assert_allclose(c.Z_n, [])

        s = c.copy()
        assert s.eq(c)

        c.change_resolution(5)
        assert c.N == 5
        c.set_coeffs(-1, None, 2)
        np.testing.assert_allclose(c.R_n, [5, 0, 0, 0, 0, 0])
        np.testing.assert_allclose(c.Z_n, [0, 0, 0, 0, 2])

        with pytest.raises(ValueError):
            c.R_n = s.R_n
        with pytest.raises(ValueError):
            c.Z_n = s.Z_n

        c.name = "my curve"
        assert "my" in c.name
        assert c.name in str(c)
        assert "FourierRZCurve" in str(c)
        assert c.sym

        c.NFP = 3
        assert c.NFP == 3
        assert c.R_basis.NFP == 3
        assert c.Z_basis.NFP == 3

    @pytest.mark.unit
    def test_asserts(self):
        """Test error checking when creating FourierRZCurve."""
        with pytest.raises(ValueError):
            _ = FourierRZCurve(R_n=[])

    @pytest.mark.unit
    def test_to_FourierXYZCurve(self):
        """Test conversion to FourierXYZCurve."""
        rz = FourierRZCurve(R_n=[0, 10, 1], Z_n=[-1, 0, 0])
        grid = LinearGrid(N=20, endpoint=False)
        xyz = rz.to_FourierXYZ(N=2, grid=grid, s=grid.nodes[:, 2])

        np.testing.assert_allclose(
            rz.compute("curvature", grid=grid)["curvature"],
            xyz.compute("curvature", grid=grid)["curvature"],
        )
        np.testing.assert_allclose(
            rz.compute("torsion", grid=grid)["torsion"],
            xyz.compute("torsion", grid=grid)["torsion"],
        )
        np.testing.assert_allclose(
            rz.compute("length", grid=grid)["length"],
            xyz.compute("length", grid=grid)["length"],
        )
        np.testing.assert_allclose(
            rz.compute("x", grid=grid, basis="xyz")["x"],
            xyz.compute("x", basis="xyz", grid=grid)["x"],
            atol=1e-12,
        )

        # same thing but with arclength angle

        xyz = rz.to_FourierXYZ(N=2, grid=grid, s="arclength")

        np.testing.assert_allclose(
            rz.compute("length", grid=grid)["length"],
            xyz.compute("length", grid=grid)["length"],
            atol=3e-3,
        )

        # pass in non-monotonic s
        grid = LinearGrid(N=20, endpoint=False)
        s = grid.nodes[:, 2]
        s[-2] = s[-1]
        with pytest.raises(ValueError):
            xyz = rz.to_FourierXYZ(N=2, grid=grid, s=s)

    @pytest.mark.unit
    def test_to_SplineXYZCurve(self):
        """Test conversion to SplineXYZCurve."""
        rz = FourierRZCurve(R_n=[0, 10, 1], Z_n=[-1, 0, 0])
        xyz = rz.to_SplineXYZ(grid=500, knots="arclength")

        grid = LinearGrid(N=20, endpoint=False)

        np.testing.assert_allclose(
            rz.compute("length", grid=grid)["length"],
            xyz.compute("length", grid=grid)["length"],
            atol=1e-2,
        )
        coords_xyz = np.asarray(xyz.compute("x", basis="rpz", grid=grid)["x"])
        phi_xyz = (coords_xyz[:, 1] + 1e-4) % (2 * np.pi)
        coords_rpz = rz.compute("x", grid=grid, basis="rpz")["x"]
        phi_rpz = (coords_rpz[:, 1] + 1e-4) % (2 * np.pi)
        np.testing.assert_allclose(
            coords_rpz[:, 0::2],
            coords_xyz[:, 0::2],
            atol=1e-1,
        )
        np.testing.assert_allclose(
            phi_rpz,
            phi_xyz,
            atol=1e-1,
        )

    @pytest.mark.unit
    def test_from_input_file(self):
        """Test getting a curve from axis guess in input file."""
        path = "tests/inputs/input.QSC_r2_5.5_desc"

        curve1 = FourierRZCurve.from_input_file(path)
        curve2 = Equilibrium(**InputReader(path).inputs[0]).axis
        curve1.change_resolution(curve2.N)

        np.testing.assert_allclose(curve1.R_n, curve2.R_n)
        np.testing.assert_allclose(curve1.Z_n, curve2.Z_n)
        np.testing.assert_allclose(curve1.NFP, curve2.NFP)
        np.testing.assert_allclose(curve1.sym, curve2.sym)

        path = "tests/inputs/input.QSC_r2_5.5_vmec"

        with pytest.warns(UserWarning):
            curve3 = FourierRZCurve.from_input_file(path)
            curve4 = Equilibrium(**InputReader(path).inputs[0]).axis
        curve3.change_resolution(curve4.N)

        np.testing.assert_allclose(curve3.R_n, curve4.R_n)
        np.testing.assert_allclose(curve3.Z_n, curve4.Z_n)
        np.testing.assert_allclose(curve3.NFP, curve4.NFP)
        np.testing.assert_allclose(curve3.sym, curve4.sym)


class TestFourierXYZCurve:
    """Tests for FourierXYZCurve class."""

    @pytest.mark.unit
    def test_length(self):
        """Test length of circular curve."""
        c = FourierXYZCurve()
        np.testing.assert_allclose(
            c.compute("length", grid=20)["length"], 2 * 2 * np.pi
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("length", grid=20)["length"], 2 * 2 * np.pi
        )

    @pytest.mark.unit
    def test_curvature(self):
        """Test curvature of circular curve."""
        c = FourierXYZCurve()
        np.testing.assert_allclose(c.compute("curvature", grid=20)["curvature"], 1 / 2)
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(c.compute("curvature", grid=20)["curvature"], 1 / 2)

    @pytest.mark.unit
    def test_torsion(self):
        """Test torsion of circular curve."""
        c = FourierXYZCurve(modes=[-1, 0, 1])
        np.testing.assert_allclose(
            c.compute("torsion", grid=20)["torsion"], 0, atol=1e-12
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("torsion", grid=20)["torsion"], 0, atol=1e-12
        )

    @pytest.mark.unit
    def test_frenet(self):
        """Test frenet-serret frame of circular curve."""
        c = FourierXYZCurve()
        data = c.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"], basis="rpz", grid=0
        )
        T, N, B = data["frenet_tangent"], data["frenet_normal"], data["frenet_binormal"]
        np.testing.assert_allclose(T, np.array([[0, 0, -1]]), atol=1e-12)
        np.testing.assert_allclose(N, np.array([[-1, 0, 0]]), atol=1e-12)
        np.testing.assert_allclose(B, np.array([[0, 1, 0]]), atol=1e-12)
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        data = c.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"], basis="xyz", grid=0
        )
        T, N, B = data["frenet_tangent"], data["frenet_normal"], data["frenet_binormal"]
        np.testing.assert_allclose(T, np.array([[0, 0, -1]]), atol=1e-12)
        np.testing.assert_allclose(N, np.array([[1, 0, 0]]), atol=1e-12)
        np.testing.assert_allclose(B, np.array([[0, 1, 0]]), atol=1e-12)

    @pytest.mark.unit
    def test_coords(self):
        """Test lab frame coordinates of circular curve."""
        c = FourierXYZCurve()
        x, y, z = c.compute("x", grid=0, basis="xyz")["x"].T
        np.testing.assert_allclose(x, 12)
        np.testing.assert_allclose(y, 0)
        np.testing.assert_allclose(z, 0)
        c.rotate(angle=np.pi / 2)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        r, p, z = c.compute("x", grid=0, basis="rpz")["x"].T
        np.testing.assert_allclose(r, np.sqrt(1**2 + 11**2))
        np.testing.assert_allclose(p, np.arctan2(-11, 1))
        np.testing.assert_allclose(z, 1)

    @pytest.mark.unit
    def test_to_FourierXYZCurve(self):
        """Test fitting FourierXYZCurve from SplineXYZCurve object."""
        npts = 4000
        # make a simple circular curve of radius 2
        R = 2
        # make initial points non-uniform in angle
        phi = 2 * np.pi * np.linspace(0, 1, 1001, endpoint=True) ** 2
        c = SplineXYZCurve(
            X=R * np.cos(phi),
            Y=R * np.sin(phi),
            Z=np.zeros_like(phi),
            knots="arclength",
        )
        c2 = c.to_FourierXYZ(N=1, grid=1000)

        np.testing.assert_allclose(
            c.compute("length", grid=npts)["length"], R * 2 * np.pi, atol=2e-3
        )
        np.testing.assert_allclose(
            c2.compute("length", grid=npts)["length"], R * 2 * np.pi, atol=2e-3
        )

        grid = LinearGrid(N=20, endpoint=False)
        coords1 = c.compute("x", grid=grid, basis="xyz")["x"]
        coords2 = c2.compute("x", grid=grid, basis="xyz")["x"]

        np.testing.assert_allclose(coords1, coords2, atol=8e-3)

    @pytest.mark.unit
    def test_misc(self):
        """Test getting/setting misc attributes of FourierXYZCurve."""
        c = FourierXYZCurve()

        X, Y, Z = c.get_coeffs(0)
        np.testing.assert_allclose(X, 10)
        np.testing.assert_allclose(Y, 0)
        np.testing.assert_allclose(Z, 0)
        c.set_coeffs(0, 5, 2, 3)
        np.testing.assert_allclose(c.X_n, [0, 5, 2])
        np.testing.assert_allclose(c.Y_n, [0, 2, 0])
        np.testing.assert_allclose(c.Z_n, [-2, 3, 0])

        s = c.copy()
        assert s.eq(c)

        c.change_resolution(5)
        assert c.N == 5
        with pytest.raises(ValueError):
            c.X_n = s.X_n
        with pytest.raises(ValueError):
            c.Y_n = s.Y_n
        with pytest.raises(ValueError):
            c.Z_n = s.Z_n


class TestPlanarCurve:
    """Tests for FourierPlanarCurve class."""

    @pytest.mark.unit
    def test_length(self):
        """Test length of circular curve."""
        c = FourierPlanarCurve(modes=[0])
        np.testing.assert_allclose(
            c.compute("length", grid=20)["length"], 2 * 2 * np.pi
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("length", grid=20)["length"], 2 * 2 * np.pi
        )

    @pytest.mark.unit
    def test_curvature(self):
        """Test curvature of circular curve."""
        c = FourierPlanarCurve()
        np.testing.assert_allclose(c.compute("curvature", grid=20)["curvature"], 1 / 2)
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(c.compute("curvature", grid=20)["curvature"], 1 / 2)

    @pytest.mark.unit
    def test_torsion(self):
        """Test torsion of circular curve."""
        c = FourierPlanarCurve()
        np.testing.assert_allclose(
            c.compute("torsion", grid=20)["torsion"], 0, atol=1e-12
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("torsion", grid=20)["torsion"], 0, atol=1e-12
        )

    @pytest.mark.unit
    def test_frenet(self):
        """Test frenet-serret frame of circular curve."""
        c = FourierPlanarCurve()
        data = c.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"], basis="xyz", grid=0
        )
        T, N, B = data["frenet_tangent"], data["frenet_normal"], data["frenet_binormal"]
        np.testing.assert_allclose(T, np.array([[0, 0, -1]]), atol=1e-12)
        np.testing.assert_allclose(N, np.array([[-1, 0, 0]]), atol=1e-12)
        np.testing.assert_allclose(B, np.array([[0, 1, 0]]), atol=1e-12)
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        data = c.compute(
            ["frenet_tangent", "frenet_normal", "frenet_binormal"], basis="xyz", grid=0
        )
        T, N, B = data["frenet_tangent"], data["frenet_normal"], data["frenet_binormal"]
        np.testing.assert_allclose(T, np.array([[0, 0, -1]]), atol=1e-12)
        np.testing.assert_allclose(N, np.array([[1, 0, 0]]), atol=1e-12)
        np.testing.assert_allclose(B, np.array([[0, 1, 0]]), atol=1e-12)

    @pytest.mark.unit
    def test_coords(self):
        """Test lab frame coordinates of circular curve."""
        c = FourierPlanarCurve()
        r, p, z = c.compute("x", grid=0, basis="rpz")["x"].T
        np.testing.assert_allclose(r, 12)
        np.testing.assert_allclose(p, 0)
        np.testing.assert_allclose(z, 0)
        dr, dp, dz = c.compute("x_sss", grid=0, basis="rpz")["x_sss"].T
        np.testing.assert_allclose(dr, 0)
        np.testing.assert_allclose(dp, 0)
        np.testing.assert_allclose(dz, 2)
        c.rotate(angle=np.pi / 2)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        x, y, z = c.compute("x", grid=0, basis="xyz")["x"].T
        np.testing.assert_allclose(x, 1)
        np.testing.assert_allclose(y, -11)
        np.testing.assert_allclose(z, 1)

    @pytest.mark.unit
    def test_misc(self):
        """Test getting/setting misc attributes of FourierPlanarCurve."""
        c = FourierPlanarCurve()

        r = c.get_coeffs(0)
        np.testing.assert_allclose(r, 2)
        c.set_coeffs(0, 3)
        np.testing.assert_allclose(c.r_n, [3])

        c.normal = [1, 2, 3]
        c.center = [3, 2, 1]
        np.testing.assert_allclose(np.linalg.norm(c.normal), 1)
        np.testing.assert_allclose(c.normal * np.linalg.norm(c.center), c.center[::-1])

        s = c.copy()
        assert s.eq(c)

        c.change_resolution(5)
        with pytest.raises(ValueError):
            c.r_n = s.r_n

    @pytest.mark.unit
    def test_asserts(self):
        """Test error checking when creating FourierPlanarCurve."""
        c = FourierPlanarCurve()
        with pytest.raises(ValueError):
            c.center = [4]
        with pytest.raises(ValueError):
            c.normal = [4]


class TestSplineXYZCurve:
    """Tests for SplineXYZCurve class."""

    @pytest.mark.unit
    def test_length(self):
        """Test length of circular curve."""
        for method in [
            "nearest",
            "linear",
            "cubic",
            "cubic2",
            "catmull-rom",
            "monotonic",
            "monotonic-0",
            "cardinal",
        ]:
            R = 1
            phi = np.linspace(0, 2 * np.pi, 1001, endpoint=True)
            # if nearest method, cant give more than the knot pts or it will return
            # a length larger than the real one
            npts = (
                2000
                if method != ["nearest", "linear"]
                else phi + 0.1 * (phi[1] - phi[0])
            )
            # make sure that length error is less than what the error would be
            # if were simply missing one segment of a linear interpolation,
            #  to try to ensure we are not making that mistake
            atol = R * 2 * np.pi / npts if method not in ["nearest", "linear"] else 3e-3
            c = SplineXYZCurve(
                X=R * np.cos(phi),
                Y=R * np.sin(phi),
                Z=np.zeros_like(phi),
                method=method,
            )
            np.testing.assert_allclose(
                c.compute("length", grid=npts)["length"],
                R * 2 * np.pi,
                atol=atol,
                err_msg=f"Failed at {method}",
            )
            c.translate([1, 1, 1])
            c.rotate(angle=np.pi)
            c.flip([0, 1, 0])
            np.testing.assert_allclose(
                c.compute("length", grid=npts)["length"],
                R * 2 * np.pi,
                atol=atol,
                err_msg=f"Failed at {method}",
            )

            # make a simple circular curve with supplied knots as phi
            phi = np.linspace(0, 2 * np.pi, 201, endpoint=False)
            c = SplineXYZCurve(
                X=R * np.cos(phi),
                Y=R * np.sin(phi),
                Z=np.zeros_like(phi),
                knots=phi,
                method=method,
            )
            np.testing.assert_allclose(
                c.compute("length", grid=npts)["length"],
                R * 2 * np.pi,
                atol=atol,
                err_msg=f"Failed at {method}",
            )
            c.translate([1, 1, 1])
            c.rotate(angle=np.pi)
            c.flip([0, 1, 0])
            np.testing.assert_allclose(
                c.compute("length", grid=npts)["length"],
                R * 2 * np.pi,
                atol=atol,
                err_msg=f"Failed at {method}",
            )

            if method == "nearest":
                continue  # don't test changing the grid if nearest
                # since it will give wrong answers for
                # grids with more than the initial num of knots
            # check lengths when changing X,Y,Z from initial values
            # and from changing grids
            R = 1.1
            c.X = R * np.cos(phi)
            c.Y = R * np.sin(phi)
            c.Z = np.ones_like(phi)
            grid = LinearGrid(zeta=np.linspace(0, 2 * np.pi, npts, endpoint=False))
            np.testing.assert_allclose(
                c.compute("length", grid=grid)["length"],
                R * 2 * np.pi,
                atol=atol,
                err_msg=f"Failed at {method}",
            )
            np.testing.assert_allclose(
                c.compute("length", grid=None)["length"],
                R * 2 * np.pi,
                atol=9e-3,
                err_msg=f"Failed at {method}",
            )

    @pytest.mark.unit
    def test_coords(self):
        """Test lab frame coordinates of circular curve."""
        # make a simple circular curve of radius 2
        R = 3
        phi = np.linspace(0, 2 * np.pi, 101, endpoint=False)
        c = SplineXYZCurve(X=R * np.cos(phi), Y=R * np.sin(phi), Z=np.zeros_like(phi))
        x, y, z = c.compute("x", grid=Grid(np.array([[0.0, 0.0, 0.0]])), basis="xyz")[
            "x"
        ].T
        np.testing.assert_allclose(x, R)
        np.testing.assert_allclose(y, 0, atol=1e-15)
        np.testing.assert_allclose(z, 0, atol=1e-15)
        c.rotate(angle=np.pi / 2)
        c.flip([0, 1, 0])
        c.translate([1, 1, 1])
        r, p, z = c.compute("x", grid=Grid(np.array([[0.0, 0.0, 0.0]])), basis="rpz")[
            "x"
        ].T
        np.testing.assert_allclose(r, np.sqrt(1**2 + (R - 1) ** 2))
        np.testing.assert_allclose(p, np.arctan2(-(R - 1), 1))
        np.testing.assert_allclose(z, 1)

    @pytest.mark.unit
    def test_curvature(self):
        """Test curvature of circular curve."""
        # make a simple circular curve of radius 10
        R = 10
        phi = np.linspace(0, 2 * np.pi, 100, endpoint=True)
        c = SplineXYZCurve(X=R * np.cos(phi), Y=R * np.sin(phi), Z=np.zeros_like(phi))
        np.testing.assert_allclose(
            c.compute("curvature", grid=10)["curvature"][1:-1], 1 / 10, atol=1e-3
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("curvature", grid=10)["curvature"][1:-1], 1 / 10, atol=1e-3
        )

    @pytest.mark.unit
    def test_torsion(self):
        """Test torsion of circular curve."""
        # make a simple circular curve of radius 10
        R = 10
        phi = np.linspace(0, 2 * np.pi, 100, endpoint=True)
        c = SplineXYZCurve(X=R * np.cos(phi), Y=R * np.sin(phi), Z=np.zeros_like(phi))
        np.testing.assert_allclose(
            c.compute("torsion", grid=20)["torsion"], 0, atol=1e-12
        )
        c.translate([1, 1, 1])
        c.rotate(angle=np.pi)
        c.flip([0, 1, 0])
        np.testing.assert_allclose(
            c.compute("torsion", grid=20)["torsion"], 0, atol=1e-12
        )

    @pytest.mark.unit
    def test_to_SplineXYZCurve(self):
        """Test converting FourierXYZCurve to SplineXYZCurve object."""
        npts = 4000
        # make a simple circular curve of radius 2
        R = 2
        c = FourierXYZCurve()
        c2 = c.to_SplineXYZ(grid=npts)

        np.testing.assert_allclose(
            c.compute("length", grid=npts)["length"], R * 2 * np.pi, atol=2e-3
        )
        np.testing.assert_allclose(
            c2.compute("length", grid=npts)["length"], R * 2 * np.pi, atol=2e-3
        )
        grid = LinearGrid(N=20, endpoint=False)
        coords1 = c.compute("x", grid=grid)["x"]
        coords2 = c2.compute("x", grid=grid)["x"]

        np.testing.assert_allclose(coords1, coords2, atol=1e-10)

    @pytest.mark.unit
    def test_asserts_and_errors(self):
        """Test error checking when creating or setting properties of SplineXYZCurve."""
        # make a simple circular curve of radius 2
        R = 2
        phi = np.linspace(0, 2 * np.pi, 101, endpoint=True)
        c = SplineXYZCurve(X=R * np.cos(phi), Y=R * np.sin(phi), Z=np.zeros_like(phi))

        # change number of knots, should raise error since is different than
        # existing knots
        phi = np.linspace(0, 2 * np.pi, 102, endpoint=True)
        with pytest.raises(ValueError):
            c.X = R * np.cos(phi)
        with pytest.raises(ValueError):
            c.Y = R * np.sin(phi)
        with pytest.raises(ValueError):
            c.Z = np.zeros_like(phi)

        # setter for knots
        with pytest.raises(ValueError):
            c.knots = np.linspace(0, 10, 10)
        knots = c.knots
        knots[-2] = knots[-1]  # make it non-monotonic
        with pytest.raises(ValueError):
            c.knots = knots
        with pytest.raises(ValueError):
            c.knots *= -1
        with pytest.raises(ValueError):
            c.knots += np.pi

        # setter for method
        with pytest.raises(ValueError):
            c.method = "not a valid method"

    @pytest.mark.unit
    def test_misc(self):
        """Test getting/setting misc attributes of SplineXYZCurve."""
        # make a simple circular curve of radius 2
        R = 2
        phi = np.linspace(0, 2 * np.pi, 101, endpoint=True)
        c = SplineXYZCurve(X=R * np.cos(phi), Y=R * np.sin(phi), Z=np.zeros_like(phi))

        s = c.copy()
        assert s.eq(c)

    @pytest.mark.unit
    def test_compute_ndarray_error(self):
        """Test raising TypeError if ndarray is passed in."""
        # make a simple circular curve of radius 2
        R = 2
        phi = np.linspace(0, 2 * np.pi, 101, endpoint=True)
        c = SplineXYZCurve(X=R * np.cos(phi), Y=R * np.sin(phi), Z=np.zeros_like(phi))
        with pytest.raises(TypeError):
            c.compute("length", grid=np.linspace(0, 1, 10))



"""Utilities for dealing with simple bound constraints."""

import functools

from desc.backend import cond, jit, jnp

from .utils import evaluate_quadratic_form_hess, evaluate_quadratic_form_jac


@jit
def cl_scaling_vector(x, g, lb, ub):
    """Compute Coleman-Li scaling vector and its derivatives.

    Components of a vector v are defined as follows:
    ::
               | ub[i] - x[i], if g[i] < 0 and ub[i] < np.inf
        v[i] = | x[i] - lb[i], if g[i] > 0 and lb[i] > -np.inf
               | 1,           otherwise

    According to this definition v[i] >= 0 for all i. It differs from the
    definition in paper [1]_ (eq. (2.2)), where the absolute value of v is
    used. Both definitions are equivalent down the line.
    Derivatives of v with respect to x take value 1, -1 or 0 depending on a
    case.

    Parameters
    ----------
    x : ndarray
        state vector
    g : ndarray
        gradient of objective function
    lb, ub : ndarray
        lower and upper bounds on x

    Returns
    -------
    v : ndarray with shape of x
        Scaling vector.
    dv : ndarray with shape of x
        Derivatives of v[i] with respect to x[i], diagonal elements of v's
        Jacobian.

    References
    ----------
    .. [1] M.A. Branch, T.F. Coleman, and Y. Li, "A Subspace, Interior,
           and Conjugate Gradient Method for Large-Scale Bound-Constrained
           Minimization Problems," SIAM Journal on Scientific Computing,
           Vol. 21, Number 1, pp 1-23, 1999.
    """
    v = jnp.ones_like(x)
    dv = jnp.zeros_like(x)

    mask = (g < 0) & jnp.isfinite(ub)
    v = jnp.where(mask, ub - x, v)
    dv = jnp.where(mask, -1, dv)

    mask = (g > 0) & jnp.isfinite(lb)
    v = jnp.where(mask, x - lb, v)
    dv = jnp.where(mask, 1, dv)

    return v, dv


@jit
def step_size_to_bound(x, s, lb, ub):
    """Compute a min_step size required to reach a bound.

    The function computes a positive scalar t, such that x + s * t is on
    the bound.

    Parameters
    ----------
    x : ndarray
        state vector
    s : ndarray
        proposed step
    lb, ub : ndarray
        lower and upper bounds on x

    Returns
    -------
    step : float
        Computed step. Non-negative value.
    hits : ndarray of int with shape of x
        Each element indicates whether a corresponding variable reaches the
        bound:
             *  0 - the bound was not hit.
             * -1 - the lower bound was hit.
             *  1 - the upper bound was hit.
    """
    steps = jnp.inf * jnp.ones(x.size)
    mask = s != 0
    steps = jnp.where(mask, jnp.maximum((lb - x) / s, (ub - x) / s), steps)
    min_step = jnp.min(steps)
    return min_step, jnp.equal(steps, min_step) * jnp.sign(s).astype(int)


@jit
def find_active_constraints(x, lb, ub, rtol=1e-10):
    """Determine which constraints are active in a given point.

    The threshold is computed using `rtol` and the absolute value of the
    closest bound.

    Parameters
    ----------
    x : ndarray
        state vector
    lb, ub : ndarray
        lower and upper bounds on x

    Returns
    -------
    active : ndarray of int with shape of x
        Each component shows whether the corresponding constraint is active:
             *  0 - a constraint is not active.
             * -1 - a lower bound is active.
             *  1 - a upper bound is active.
    """

    def truefun(active):
        active = jnp.where(x <= lb, -1, active)
        active = jnp.where(x >= ub, 1, active)
        return active

    def falsefun(active):
        lower_dist = x - lb
        upper_dist = ub - x

        lower_threshold = rtol * jnp.maximum(1, jnp.abs(lb))
        upper_threshold = rtol * jnp.maximum(1, jnp.abs(ub))

        lower_active = jnp.isfinite(lb) & (
            lower_dist <= jnp.minimum(upper_dist, lower_threshold)
        )
        active = jnp.where(lower_active, -1, active)

        upper_active = jnp.isfinite(ub) & (
            upper_dist <= jnp.minimum(lower_dist, upper_threshold)
        )
        active = jnp.where(upper_active, 1, active)

        return active

    return cond(rtol == 0, truefun, falsefun, jnp.zeros_like(x, dtype=int))


@jit
def make_strictly_feasible(x, lb, ub, rstep=1e-10):
    """Shift a point to the interior of a feasible region.

    Each element of the returned vector is at least at a relative distance
    `rstep` from the closest bound. If ``rstep=0`` then `np.nextafter` is used.
    """
    active = find_active_constraints(x, lb, ub, rstep)
    lower_mask = jnp.equal(active, -1)
    upper_mask = jnp.equal(active, 1)

    lwr = jnp.where(
        rstep == 0, jnp.nextafter(lb, ub), lb + rstep * jnp.maximum(1, jnp.abs(lb))
    )
    upr = jnp.where(
        rstep == 0, jnp.nextafter(ub, lb), ub - rstep * jnp.maximum(1, jnp.abs(ub))
    )

    x_new = x.copy()
    x_new = jnp.where(lower_mask, lwr, x_new)
    x_new = jnp.where(upper_mask, upr, x_new)
    tight_bounds = (x_new < lb) | (x_new > ub)
    x_new = jnp.where(tight_bounds, 0.5 * (lb + ub), x_new)

    return x_new


@jit
def in_bounds(x, lb, ub):
    """Check if a point lies within bounds."""
    return jnp.all((x >= lb) & (x <= ub))


@functools.partial(jit, static_argnames="mode")
def select_step(x, JorH, diag_h, g_h, p, p_h, d, Delta, lb, ub, theta, mode="jac"):
    """Select the best step according to Trust Region Reflective algorithm.

    The algorithm chooses the best step among three options:

        - a constrained trust-region step (p)
        - a reflected step (r)
        - a constrained Cauchy step (ag, minimizer along direction of scaled gradient)
    """
    assert mode in ["jac", "hess"]

    if mode == "jac":
        evaluate_quadratic_form = evaluate_quadratic_form_jac
        build_quadratic_1d = build_quadratic_1d_jac
    else:
        evaluate_quadratic_form = evaluate_quadratic_form_hess
        build_quadratic_1d = build_quadratic_1d_hess

    def inbounds_true(p, p_h):
        p_value = evaluate_quadratic_form(JorH, g_h, p_h, diag=diag_h)
        return p, p_h, -p_value

    def inbounds_false(p, p_h):
        p_stride, hits = step_size_to_bound(x, p, lb, ub)

        # Compute the reflected direction.
        r_h = jnp.copy(p_h)
        r_h = jnp.where(hits.astype(bool), r_h * -1, r_h)
        r = d * r_h

        # Restrict trust-region step, such that it hits the bound.
        p *= p_stride
        p_h *= p_stride
        x_on_bound = x + p

        # Reflected direction will cross first either feasible region or trust region
        # boundary.
        _, to_tr = intersect_trust_region(p_h, r_h, Delta)
        to_bound, _ = step_size_to_bound(x_on_bound, r, lb, ub)

        # Find lower and upper bounds on a step size along the reflected
        # direction, considering the strict feasibility requirement. There is no
        # single correct way to do that, the chosen approach seems to work best
        # on test problems.
        r_stride = jnp.minimum(to_bound, to_tr)
        r_stride_l = cond(
            r_stride > 0,
            lambda _: (1 - theta) * p_stride / r_stride,
            lambda _: 0.0,
            None,
        )
        r_stride_u = cond(
            r_stride > 0,
            lambda _: jnp.where(r_stride == to_bound, theta * to_bound, to_tr),
            lambda _: -1.0,
            None,
        )

        # Check if reflection step is available.
        def r_truefun(r_h, r):
            a, b, c = build_quadratic_1d(JorH, g_h, r_h, s0=p_h, diag=diag_h)
            r_stride, r_value = minimize_quadratic_1d(a, b, r_stride_l, r_stride_u, c=c)
            r_h *= r_stride
            r_h += p_h
            r = r_h * d
            return r_h, r, r_value

        def r_falsefun(r_h, r):
            return r_h, r, jnp.inf

        r_h, r, r_value = cond(r_stride_l <= r_stride_u, r_truefun, r_falsefun, r_h, r)

        # Now correct p_h to make it strictly interior.
        p *= theta
        p_h *= theta
        p_value = evaluate_quadratic_form(JorH, g_h, p_h, diag=diag_h)

        ag_h = -g_h
        ag = d * ag_h

        to_tr = Delta / jnp.linalg.norm(ag_h)
        to_bound, _ = step_size_to_bound(x, ag, lb, ub)
        ag_stride = jnp.where(to_bound < to_tr, theta * to_bound, to_tr)

        a, b = build_quadratic_1d(JorH, g_h, ag_h, diag=diag_h)
        ag_stride, ag_value = minimize_quadratic_1d(a, b, 0, ag_stride)
        ag_h *= ag_stride
        ag *= ag_stride

        values = jnp.array([p_value, r_value, ag_value])
        steps = jnp.array([p, r, ag])
        steps_h = jnp.array([p_h, r_h, ag_h])
        idx = jnp.nanargmin(values)
        return steps[idx], steps_h[idx], -values[idx]

    return cond(in_bounds(x + p, lb, ub), inbounds_true, inbounds_false, p, p_h)


@jit
def build_quadratic_1d_jac(J, g, s, diag=None, s0=None):
    """Parameterize a multivariate quadratic function along a line.

    The resulting univariate quadratic function is given as follows:
    ::
        f(t) = 0.5 * (s0 + s*t).T * (J.T*J + diag) * (s0 + s*t) +
               g.T * (s0 + s*t)

    Parameters
    ----------
    J : ndarray, shape (m, n)
        Jacobian matrix, affects the quadratic term.
    g : ndarray, shape (n,)
        Gradient, defines the linear term.
    s : ndarray, shape (n,)
        Direction vector of a line.
    diag : None or ndarray with shape (n,), optional
        Addition diagonal part, affects the quadratic term.
        If None, assumed to be 0.
    s0 : None or ndarray with shape (n,), optional
        Initial point. If None, assumed to be 0.

    Returns
    -------
    a : float
        Coefficient for t**2.
    b : float
        Coefficient for t.
    c : float
        Free term. Returned only if `s0` is provided.
    """
    v = J.dot(s)
    a = jnp.dot(v, v)
    if diag is not None:
        a += jnp.dot(s * diag, s)
    a *= 0.5

    b = jnp.dot(g, s)

    if s0 is not None:
        u = J.dot(s0)
        b += jnp.dot(u, v)
        c = 0.5 * jnp.dot(u, u) + jnp.dot(g, s0)
        if diag is not None:
            b += jnp.dot(s0 * diag, s)
            c += 0.5 * jnp.dot(s0 * diag, s0)
        return a, b, c
    else:
        return a, b


@jit
def build_quadratic_1d_hess(H, g, s, diag=None, s0=None):
    """Parameterize a multivariate quadratic function along a line.

    The resulting univariate quadratic function is given as follows:
    ::
        f(t) = 0.5 * (s0 + s*t).T * (H + diag) * (s0 + s*t) +
               g.T * (s0 + s*t)

    Parameters
    ----------
    H : ndarray, shape (n, n)
        Hessian matrix, affects the quadratic term.
    g : ndarray, shape (n,)
        Gradient, defines the linear term.
    s : ndarray, shape (n,)
        Direction vector of a line.
    diag : None or ndarray with shape (n,), optional
        Addition diagonal part, affects the quadratic term.
        If None, assumed to be 0.
    s0 : None or ndarray with shape (n,), optional
        Initial point. If None, assumed to be 0.

    Returns
    -------
    a : float
        Coefficient for t**2.
    b : float
        Coefficient for t.
    c : float
        Free term. Returned only if `s0` is provided.
    """
    a = H.dot(s).dot(s)
    if diag is not None:
        a += jnp.dot(s * diag, s)
    a *= 0.5

    b = jnp.dot(g, s)

    if s0 is not None:
        u = H.dot(s0)
        b += jnp.dot(u, s)
        c = 0.5 * jnp.dot(u, s0) + jnp.dot(g, s0)
        if diag is not None:
            b += jnp.dot(s0 * diag, s)
            c += 0.5 * jnp.dot(s0 * diag, s0)
        return a, b, c
    else:
        return a, b


@jit
def minimize_quadratic_1d(a, b, lb, ub, c=0):
    """Minimize a 1-D quadratic function subject to bounds.

    The free term `c` is 0 by default. Bounds must be finite.

    Returns
    -------
    t : float
        Minimum point.
    y : float
        Minimum value.
    """
    t = [lb, ub]
    extremum = -0.5 * b / a
    # need to split double sided inequality here to work with jax
    t.append(jnp.where((lb < extremum) & (extremum < ub), extremum, jnp.nan))
    t = jnp.asarray(t)
    y = t * (a * t + b) + c
    min_index = jnp.nanargmin(y)
    return t[min_index], y[min_index]


@jit
def intersect_trust_region(x, s, Delta):
    """Find the intersection of a line with the boundary of a trust region.

    This function solves the quadratic equation with respect to t

    ||(x + s*t)||**2 = Delta**2.

    Returns
    -------
    t_neg, t_pos : tuple of float
        Negative and positive roots.

    """
    a = jnp.dot(s, s)
    b = jnp.dot(x, s)
    # c should always be <= 0, but sometimes rounding error makes it positive
    # so we clip it
    c = jnp.minimum(jnp.dot(x, x) - Delta**2, 0)
    d = jnp.sqrt(b * b - a * c)  # Root from one fourth of the discriminant.

    q = -(b + jnp.sign(b) * jnp.abs(d))
    t1 = q / a
    t2 = c / q
    t1, t2 = jnp.sort(jnp.array([t1, t2]))
    # one root should be positive, one negative, but rounding error
    # can throw that off, so we clip
    return jnp.minimum(t1, 0), jnp.maximum(t2, 0)
